from judgeval.scorers.api_scorer import APIScorerConfig
from judgeval.constants import APIScorerType
from judgeval.data import ExampleParams
from typing import List


class AnswerRelevancyScorer(APIScorerConfig):
    score_type: APIScorerType = APIScorerType.ANSWER_RELEVANCY
    required_params: List[ExampleParams] = [
        ExampleParams.INPUT,
        ExampleParams.ACTUAL_OUTPUT,
    ]



from __future__ import annotations
import functools
from typing import (
    Tuple,
    Optional,
    Any,
    TYPE_CHECKING,
    Union,
    AsyncGenerator,
    Generator,
    Iterator,
    AsyncIterator,
)
from functools import wraps
from enum import Enum
from judgeval.data.trace import TraceUsage
from judgeval.logger import judgeval_logger
from litellm.cost_calculator import cost_per_token as _original_cost_per_token
from opentelemetry.trace import Span

from judgeval.tracer.llm.providers import (
    HAS_OPENAI,
    HAS_TOGETHER,
    HAS_ANTHROPIC,
    HAS_GOOGLE_GENAI,
    HAS_GROQ,
    ApiClient,
)
from judgeval.tracer.managers import sync_span_context, async_span_context
from judgeval.tracer.keys import AttributeKeys
from judgeval.utils.serialize import safe_serialize
from judgeval.tracer.utils import set_span_attribute

if TYPE_CHECKING:
    from judgeval.tracer import Tracer


class ProviderType(Enum):
    """Enum for different LLM provider types."""

    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    TOGETHER = "together"
    GOOGLE = "google"
    GROQ = "groq"
    DEFAULT = "default"


@wraps(_original_cost_per_token)
def cost_per_token(
    *args: Any, **kwargs: Any
) -> Tuple[Optional[float], Optional[float]]:
    try:
        prompt_tokens_cost_usd_dollar, completion_tokens_cost_usd_dollar = (
            _original_cost_per_token(*args, **kwargs)
        )
        if (
            prompt_tokens_cost_usd_dollar == 0
            and completion_tokens_cost_usd_dollar == 0
        ):
            judgeval_logger.warning("LiteLLM returned a total of 0 for cost per token")
        return prompt_tokens_cost_usd_dollar, completion_tokens_cost_usd_dollar
    except Exception as e:
        judgeval_logger.warning(f"Error calculating cost per token: {e}")
        return None, None


def _detect_provider(client: ApiClient) -> ProviderType:
    """Detect the provider type of the client once to avoid repeated isinstance checks."""
    if HAS_OPENAI:
        from judgeval.tracer.llm.providers import openai_OpenAI, openai_AsyncOpenAI

        assert openai_OpenAI is not None, "OpenAI client not found"
        assert openai_AsyncOpenAI is not None, "OpenAI async client not found"
        if isinstance(client, (openai_OpenAI, openai_AsyncOpenAI)):
            return ProviderType.OPENAI

    if HAS_ANTHROPIC:
        from judgeval.tracer.llm.providers import (
            anthropic_Anthropic,
            anthropic_AsyncAnthropic,
        )

        assert anthropic_Anthropic is not None, "Anthropic client not found"
        assert anthropic_AsyncAnthropic is not None, "Anthropic async client not found"
        if isinstance(client, (anthropic_Anthropic, anthropic_AsyncAnthropic)):
            return ProviderType.ANTHROPIC

    if HAS_TOGETHER:
        from judgeval.tracer.llm.providers import (
            together_Together,
            together_AsyncTogether,
        )

        assert together_Together is not None, "Together client not found"
        assert together_AsyncTogether is not None, "Together async client not found"
        if isinstance(client, (together_Together, together_AsyncTogether)):
            return ProviderType.TOGETHER

    if HAS_GOOGLE_GENAI:
        from judgeval.tracer.llm.providers import (
            google_genai_Client,
            google_genai_AsyncClient,
        )

        assert google_genai_Client is not None, "Google GenAI client not found"
        assert google_genai_AsyncClient is not None, (
            "Google GenAI async client not found"
        )
        if isinstance(client, (google_genai_Client, google_genai_AsyncClient)):
            return ProviderType.GOOGLE

    if HAS_GROQ:
        from judgeval.tracer.llm.providers import groq_Groq, groq_AsyncGroq

        assert groq_Groq is not None, "Groq client not found"
        assert groq_AsyncGroq is not None, "Groq async client not found"
        if isinstance(client, (groq_Groq, groq_AsyncGroq)):
            return ProviderType.GROQ

    return ProviderType.DEFAULT


# Provider-specific content extraction handlers
def _extract_openai_content(chunk) -> str:
    """Extract content from OpenAI streaming chunk."""
    if (
        hasattr(chunk, "choices")
        and chunk.choices
        and hasattr(chunk.choices[0], "delta")
    ):
        delta_content = getattr(chunk.choices[0].delta, "content", None)
        if delta_content:
            return delta_content
    return ""


def _extract_anthropic_content(chunk) -> str:
    """Extract content from Anthropic streaming chunk."""
    if hasattr(chunk, "type") and chunk.type == "content_block_delta":
        if hasattr(chunk, "delta") and hasattr(chunk.delta, "text"):
            return chunk.delta.text or ""
    elif hasattr(chunk, "delta") and hasattr(chunk.delta, "text"):
        return chunk.delta.text or ""
    elif hasattr(chunk, "text"):
        return chunk.text or ""
    return ""


def _extract_together_content(chunk) -> str:
    """Extract content from Together streaming chunk."""
    if hasattr(chunk, "choices") and chunk.choices:
        choice = chunk.choices[0]
        if hasattr(choice, "delta") and hasattr(choice.delta, "content"):
            return choice.delta.content or ""
    return ""


def _extract_groq_content(chunk) -> str:
    """Extract content from Groq streaming chunk."""
    if hasattr(chunk, "choices") and chunk.choices:
        choice = chunk.choices[0]
        if hasattr(choice, "delta") and hasattr(choice.delta, "content"):
            return choice.delta.content or ""
    return ""


# Provider-specific chunk usage extraction handlers
def _extract_openai_chunk_usage(chunk) -> Any:
    """Extract usage data from OpenAI streaming chunk."""
    if hasattr(chunk, "usage") and chunk.usage:
        return chunk.usage
    return None


def _extract_anthropic_chunk_usage(chunk) -> Any:
    """Extract usage data from Anthropic streaming chunk."""
    if hasattr(chunk, "type"):
        if chunk.type == "message_start":
            if hasattr(chunk, "message") and hasattr(chunk.message, "usage"):
                return chunk.message.usage
        elif chunk.type == "message_delta":
            if hasattr(chunk, "usage"):
                return chunk.usage
        elif chunk.type == "message_stop":
            if hasattr(chunk, "usage"):
                return chunk.usage
    return None


def _extract_together_chunk_usage(chunk) -> Any:
    """Extract usage data from Together streaming chunk."""
    if hasattr(chunk, "usage") and chunk.usage:
        return chunk.usage
    return None


def _extract_groq_chunk_usage(chunk) -> Any:
    """Extract usage data from Groq streaming chunk."""
    # Groq provides usage data in the last chunk when stream_options={"include_usage": True} is used
    if hasattr(chunk, "usage") and chunk.usage:
        return chunk.usage
    return None


# Provider-specific token extraction handlers
def _extract_openai_tokens(usage_data) -> tuple[int, int, int, int]:
    """Extract token counts from OpenAI usage data."""
    prompt_tokens = (
        usage_data.prompt_tokens
        if hasattr(usage_data, "prompt_tokens") and usage_data.prompt_tokens is not None
        else 0
    )
    completion_tokens = (
        usage_data.completion_tokens
        if hasattr(usage_data, "completion_tokens")
        and usage_data.completion_tokens is not None
        else 0
    )
    return prompt_tokens, completion_tokens, 0, 0


def _extract_anthropic_tokens(usage_data) -> tuple[int, int, int, int]:
    """Extract token counts from Anthropic usage data."""
    prompt_tokens = (
        usage_data.input_tokens
        if hasattr(usage_data, "input_tokens") and usage_data.input_tokens is not None
        else 0
    )
    completion_tokens = (
        usage_data.output_tokens
        if hasattr(usage_data, "output_tokens") and usage_data.output_tokens is not None
        else 0
    )
    cache_read_input_tokens = (
        usage_data.cache_read_input_tokens
        if hasattr(usage_data, "cache_read_input_tokens")
        and usage_data.cache_read_input_tokens is not None
        else 0
    )
    cache_creation_input_tokens = (
        usage_data.cache_creation_input_tokens
        if hasattr(usage_data, "cache_creation_input_tokens")
        and usage_data.cache_creation_input_tokens is not None
        else 0
    )
    return (
        prompt_tokens,
        completion_tokens,
        cache_read_input_tokens,
        cache_creation_input_tokens,
    )


def _extract_together_tokens(usage_data) -> tuple[int, int, int, int]:
    """Extract token counts from Together usage data."""
    prompt_tokens = (
        usage_data.prompt_tokens
        if hasattr(usage_data, "prompt_tokens") and usage_data.prompt_tokens is not None
        else 0
    )
    completion_tokens = (
        usage_data.completion_tokens
        if hasattr(usage_data, "completion_tokens")
        and usage_data.completion_tokens is not None
        else 0
    )
    return prompt_tokens, completion_tokens, 0, 0


def _extract_groq_tokens(usage_data) -> tuple[int, int, int, int]:
    """Extract token counts from Groq usage data."""
    prompt_tokens = (
        usage_data.prompt_tokens
        if hasattr(usage_data, "prompt_tokens") and usage_data.prompt_tokens is not None
        else 0
    )
    completion_tokens = (
        usage_data.completion_tokens
        if hasattr(usage_data, "completion_tokens")
        and usage_data.completion_tokens is not None
        else 0
    )
    # Extract cached tokens from prompt_tokens_details.cached_tokens
    cache_read_input_tokens = 0
    if (
        hasattr(usage_data, "prompt_tokens_details")
        and usage_data.prompt_tokens_details
    ):
        if (
            hasattr(usage_data.prompt_tokens_details, "cached_tokens")
            and usage_data.prompt_tokens_details.cached_tokens is not None
        ):
            cache_read_input_tokens = usage_data.prompt_tokens_details.cached_tokens

    return prompt_tokens, completion_tokens, cache_read_input_tokens, 0


# Provider-specific output formatting handlers
def _format_openai_output(response: Any) -> tuple[Optional[str], Optional[TraceUsage]]:
    """Format output data from OpenAI response."""
    from judgeval.tracer.llm.providers import (
        openai_ChatCompletion,
        openai_Response,
        openai_ParsedChatCompletion,
    )

    model_name = None
    message_content = None
    prompt_tokens = 0
    completion_tokens = 0
    cache_read_input_tokens = 0
    cache_creation_input_tokens = 0

    if openai_ChatCompletion and isinstance(response, openai_ChatCompletion):
        model_name = response.model or ""
        prompt_tokens = (
            response.usage.prompt_tokens
            if response.usage and response.usage.prompt_tokens is not None
            else 0
        )
        completion_tokens = (
            response.usage.completion_tokens
            if response.usage and response.usage.completion_tokens is not None
            else 0
        )
        cache_read_input_tokens = (
            response.usage.prompt_tokens_details.cached_tokens
            if response.usage
            and response.usage.prompt_tokens_details
            and response.usage.prompt_tokens_details.cached_tokens is not None
            else 0
        )

        if openai_ParsedChatCompletion and isinstance(
            response, openai_ParsedChatCompletion
        ):
            message_content = response.choices[0].message.parsed
        else:
            message_content = response.choices[0].message.content
    elif openai_Response and isinstance(response, openai_Response):
        model_name = response.model or ""
        prompt_tokens = (
            response.usage.input_tokens
            if response.usage and response.usage.input_tokens is not None
            else 0
        )
        completion_tokens = (
            response.usage.output_tokens
            if response.usage and response.usage.output_tokens is not None
            else 0
        )
        cache_read_input_tokens = (
            response.usage.input_tokens_details.cached_tokens
            if response.usage
            and response.usage.input_tokens_details
            and response.usage.input_tokens_details.cached_tokens is not None
            else 0
        )
        output0 = response.output[0]
        if (
            hasattr(output0, "content")
            and output0.content
            and hasattr(output0.content, "__iter__")
        ):
            message_content = "".join(
                seg.text for seg in output0.content if hasattr(seg, "text") and seg.text
            )

    if model_name:
        return message_content, _create_usage(
            model_name,
            prompt_tokens,
            completion_tokens,
            cache_read_input_tokens,
            cache_creation_input_tokens,
        )

    return None, None


def _format_anthropic_output(
    response: Any,
) -> tuple[Optional[str], Optional[TraceUsage]]:
    """Format output data from Anthropic response."""
    model_name = getattr(response, "model", "") or ""
    usage = getattr(response, "usage", None)
    prompt_tokens = (
        usage.input_tokens
        if usage and hasattr(usage, "input_tokens") and usage.input_tokens is not None
        else 0
    )
    completion_tokens = (
        usage.output_tokens
        if usage and hasattr(usage, "output_tokens") and usage.output_tokens is not None
        else 0
    )
    cache_read_input_tokens = (
        usage.cache_read_input_tokens
        if usage
        and hasattr(usage, "cache_read_input_tokens")
        and usage.cache_read_input_tokens is not None
        else 0
    )
    cache_creation_input_tokens = (
        usage.cache_creation_input_tokens
        if usage
        and hasattr(usage, "cache_creation_input_tokens")
        and usage.cache_creation_input_tokens is not None
        else 0
    )
    message_content = response.content[0].text if hasattr(response, "content") else None

    if model_name:
        return message_content, _create_usage(
            model_name,
            prompt_tokens,
            completion_tokens,
            cache_read_input_tokens,
            cache_creation_input_tokens,
        )

    return None, None


def _format_together_output(
    response: Any,
) -> tuple[Optional[str], Optional[TraceUsage]]:
    """Format output data from Together response."""
    model_name = (response.model or "") if hasattr(response, "model") else ""
    prompt_tokens = (
        response.usage.prompt_tokens
        if hasattr(response.usage, "prompt_tokens")
        and response.usage.prompt_tokens is not None
        else 0
    )
    completion_tokens = (
        response.usage.completion_tokens
        if hasattr(response.usage, "completion_tokens")
        and response.usage.completion_tokens is not None
        else 0
    )
    message_content = (
        response.choices[0].message.content if hasattr(response, "choices") else None
    )

    if model_name:
        model_name = "together_ai/" + model_name
        return message_content, _create_usage(
            model_name,
            prompt_tokens,
            completion_tokens,
            0,
            0,
        )

    return None, None


def _format_google_output(response: Any) -> tuple[Optional[str], Optional[TraceUsage]]:
    """Format output data from Google GenAI response."""
    model_name = getattr(response, "model_version", "") or ""
    usage_metadata = getattr(response, "usage_metadata", None)
    prompt_tokens = (
        usage_metadata.prompt_token_count
        if usage_metadata
        and hasattr(usage_metadata, "prompt_token_count")
        and usage_metadata.prompt_token_count is not None
        else 0
    )
    completion_tokens = (
        usage_metadata.candidates_token_count
        if usage_metadata
        and hasattr(usage_metadata, "candidates_token_count")
        and usage_metadata.candidates_token_count is not None
        else 0
    )
    message_content = (
        response.candidates[0].content.parts[0].text
        if hasattr(response, "candidates")
        else None
    )

    cache_read_input_tokens = 0
    if usage_metadata and hasattr(usage_metadata, "cached_content_token_count"):
        cache_read_input_tokens = usage_metadata.cached_content_token_count or 0

    if model_name:
        return message_content, _create_usage(
            model_name,
            prompt_tokens,
            completion_tokens,
            cache_read_input_tokens,
            0,
        )

    return None, None


def _format_groq_output(response: Any) -> tuple[Optional[str], Optional[TraceUsage]]:
    """Format output data from Groq response."""
    model_name = (response.model or "") if hasattr(response, "model") else ""
    prompt_tokens = (
        response.usage.prompt_tokens
        if hasattr(response.usage, "prompt_tokens")
        and response.usage.prompt_tokens is not None
        else 0
    )
    completion_tokens = (
        response.usage.completion_tokens
        if hasattr(response.usage, "completion_tokens")
        and response.usage.completion_tokens is not None
        else 0
    )
    # Extract cached tokens from prompt_tokens_details.cached_tokens
    cache_read_input_tokens = 0
    if (
        hasattr(response, "usage")
        and response.usage
        and hasattr(response.usage, "prompt_tokens_details")
        and response.usage.prompt_tokens_details
    ):
        if (
            hasattr(response.usage.prompt_tokens_details, "cached_tokens")
            and response.usage.prompt_tokens_details.cached_tokens is not None
        ):
            cache_read_input_tokens = response.usage.prompt_tokens_details.cached_tokens

    message_content = (
        response.choices[0].message.content if hasattr(response, "choices") else None
    )

    if model_name:
        model_name = "groq/" + model_name
        return message_content, _create_usage(
            model_name,
            prompt_tokens,
            completion_tokens,
            cache_read_input_tokens,
            0,
        )

    return None, None


class _TracedGeneratorBase:
    """Base class with common logic for parsing stream chunks."""

    __slots__ = (
        "tracer",
        "client",
        "span",
        "accumulated_content",
        "model_name",
        "provider_type",
    )

    tracer: Tracer
    client: ApiClient
    span: Span
    accumulated_content: str
    model_name: str
    provider_type: ProviderType

    def __init__(self, tracer: Tracer, client: ApiClient, span: Span, model_name: str):
        """Initialize the base traced generator.

        Args:
            tracer: The tracer instance
            client: The API client
            span: The OpenTelemetry span
            model_name: The model name (empty string default allows fallback to usage_data.model)
        """
        self.tracer = tracer
        self.client = client
        self.span = span
        self.accumulated_content = ""
        self.model_name = model_name
        self.provider_type = _detect_provider(client)

    def _extract_content(self, chunk) -> str:
        """Extract content from streaming chunk based on provider."""
        if self.provider_type == ProviderType.OPENAI:
            return _extract_openai_content(chunk)
        elif self.provider_type == ProviderType.ANTHROPIC:
            return _extract_anthropic_content(chunk)
        elif self.provider_type == ProviderType.TOGETHER:
            return _extract_together_content(chunk)
        elif self.provider_type == ProviderType.GROQ:
            return _extract_groq_content(chunk)
        else:
            # Default case - assume OpenAI-compatible for unknown providers
            return _extract_openai_content(chunk)

    def _process_chunk_usage(self, chunk):
        """Process usage data from streaming chunks based on provider."""
        usage_data = _extract_chunk_usage(self.client, chunk)
        if usage_data:
            _process_usage_data(
                self.span, usage_data, self.tracer, self.client, self.model_name
            )

    def __del__(self):
        """
        Fallback cleanup for unclosed spans. This is a safety mechanism only - spans
        should normally be finalized in StopIteration/StopAsyncIteration handlers.

        Note: __del__ is not guaranteed to be called in all situations (e.g., reference
        cycles, program exit), so this should not be relied upon as the primary cleanup
        mechanism. The primary finalization happens in the iterator protocol methods.
        """
        if self.span:
            try:
                self._finalize_span()
            except Exception as e:
                judgeval_logger.warning(
                    f"Error during span finalization in __del__: {e}"
                )

    def _finalize_span(self):
        """Finalize the span by setting completion content and ending it."""
        if self.span:
            set_span_attribute(
                self.span, AttributeKeys.GEN_AI_COMPLETION, self.accumulated_content
            )
            self.span.end()
            self.span = None


class TracedGenerator(_TracedGeneratorBase):
    """Generator wrapper that adds OpenTelemetry tracing without consuming the stream."""

    __slots__ = ("generator",)

    generator: Union[Generator[Any, None, None], Iterator[Any]]

    def __init__(
        self,
        tracer: Tracer,
        generator: Union[Generator[Any, None, None], Iterator[Any]],
        client: ApiClient,
        span: Span,
        model_name: str,
    ):
        super().__init__(tracer, client, span, model_name)
        self.generator = generator

    def __iter__(self):
        return self

    def __next__(self):
        try:
            chunk = next(self.generator)

            content = self._extract_content(chunk)
            if content:
                self.accumulated_content += content
            self._process_chunk_usage(chunk)

            return chunk

        except StopIteration:
            self._finalize_span()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise


class TracedAsyncGenerator(_TracedGeneratorBase):
    """Async generator wrapper that adds OpenTelemetry tracing without consuming the stream."""

    __slots__ = ("async_generator",)

    async_generator: Union[AsyncGenerator[Any, None], AsyncIterator[Any]]

    def __init__(
        self,
        tracer: Tracer,
        async_generator: Union[AsyncGenerator[Any, None], AsyncIterator[Any]],
        client: ApiClient,
        span: Span,
        model_name: str,
    ):
        super().__init__(tracer, client, span, model_name)
        self.async_generator = async_generator

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            chunk = await self.async_generator.__anext__()

            content = self._extract_content(chunk)
            if content:
                self.accumulated_content += content

            self._process_chunk_usage(chunk)

            return chunk

        except StopAsyncIteration:
            self._finalize_span()
            raise
        except Exception as e:
            if self.span:
                self.span.record_exception(e)
                self.span.end()
            raise


class TracedSyncContextManager:
    """Sync context manager wrapper for streaming methods."""

    def __init__(
        self,
        tracer: Tracer,
        context_manager: Any,
        client: ApiClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.context_manager = context_manager
        self.client = client
        self.span = span
        self.stream: Optional[Any] = None
        self.model_name = model_name

    def __enter__(self):
        self.stream = self.context_manager.__enter__()
        return TracedGenerator(
            self.tracer, self.stream, self.client, self.span, self.model_name
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.context_manager.__exit__(exc_type, exc_val, exc_tb)

    def __del__(self):
        """Cleanup span if not properly closed."""
        if self.span:
            try:
                self.span.end()
            except Exception:
                pass


class TracedAsyncContextManager:
    """Async context manager wrapper for streaming methods."""

    def __init__(
        self,
        tracer: Tracer,
        context_manager: Any,
        client: ApiClient,
        span: Span,
        model_name: str,
    ):
        self.tracer = tracer
        self.context_manager = context_manager
        self.client = client
        self.span = span
        self.stream: Optional[Any] = None
        self.model_name = model_name

    async def __aenter__(self):
        self.stream = await self.context_manager.__aenter__()
        return TracedAsyncGenerator(
            self.tracer, self.stream, self.client, self.span, self.model_name
        )

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.context_manager.__aexit__(exc_type, exc_val, exc_tb)

    def __del__(self):
        """Cleanup span if not properly closed."""
        if self.span:
            try:
                self.span.end()
            except Exception:
                pass


def _extract_chunk_usage(client: ApiClient, chunk) -> Any:
    """Extract usage data from streaming chunks based on provider."""
    provider_type = _detect_provider(client)

    if provider_type == ProviderType.OPENAI:
        return _extract_openai_chunk_usage(chunk)
    elif provider_type == ProviderType.ANTHROPIC:
        return _extract_anthropic_chunk_usage(chunk)
    elif provider_type == ProviderType.TOGETHER:
        return _extract_together_chunk_usage(chunk)
    elif provider_type == ProviderType.GROQ:
        return _extract_groq_chunk_usage(chunk)
    else:
        # Default case - assume OpenAI-compatible for unknown providers
        return _extract_openai_chunk_usage(chunk)


def _extract_usage_tokens(client: ApiClient, usage_data) -> tuple[int, int, int, int]:
    """Extract token counts from usage data based on provider."""
    provider_type = _detect_provider(client)

    if provider_type == ProviderType.OPENAI:
        return _extract_openai_tokens(usage_data)
    elif provider_type == ProviderType.ANTHROPIC:
        return _extract_anthropic_tokens(usage_data)
    elif provider_type == ProviderType.TOGETHER:
        return _extract_together_tokens(usage_data)
    elif provider_type == ProviderType.GROQ:
        return _extract_groq_tokens(usage_data)
    else:
        # Default case - assume OpenAI-compatible for unknown providers
        return _extract_openai_tokens(usage_data)


def _process_usage_data(
    span, usage_data, tracer: Tracer, client: ApiClient, model_name: str
):
    """Process usage data and set span attributes."""
    (
        prompt_tokens,
        completion_tokens,
        cache_read_input_tokens,
        cache_creation_input_tokens,
    ) = _extract_usage_tokens(client, usage_data)

    if prompt_tokens or completion_tokens:
        final_model_name = getattr(usage_data, "model", None) or model_name

        # Add provider prefixes for cost calculation
        provider_type = _detect_provider(client)
        if (
            provider_type == ProviderType.TOGETHER
            and final_model_name
            and not final_model_name.startswith("together_ai/")
        ):
            final_model_name = "together_ai/" + final_model_name
        elif (
            provider_type == ProviderType.GROQ
            and final_model_name
            and not final_model_name.startswith("groq/")
        ):
            final_model_name = "groq/" + final_model_name

        usage = _create_usage(
            final_model_name,
            prompt_tokens,
            completion_tokens,
            cache_read_input_tokens,
            cache_creation_input_tokens,
        )
        _set_usage_attributes(span, usage, tracer)


def _set_usage_attributes(span, usage: TraceUsage, tracer: Tracer):
    """Set usage attributes on the span for non-streaming responses."""

    set_span_attribute(span, AttributeKeys.GEN_AI_RESPONSE_MODEL, usage.model_name)
    set_span_attribute(
        span, AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS, usage.prompt_tokens
    )
    set_span_attribute(
        span, AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS, usage.completion_tokens
    )
    set_span_attribute(
        span, AttributeKeys.GEN_AI_USAGE_COMPLETION_TOKENS, usage.completion_tokens
    )
    set_span_attribute(
        span, AttributeKeys.GEN_AI_USAGE_TOTAL_COST, usage.total_cost_usd
    )
    tracer.add_cost_to_current_context(usage.total_cost_usd)


def wrap_provider(tracer: Tracer, client: ApiClient) -> ApiClient:
    """
    Wraps an API client to add tracing capabilities.
    Supports OpenAI, Together, Anthropic, Google GenAI, and Groq clients.
    """

    def wrapped(function, span_name):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                span = tracer.get_tracer().start_span(
                    span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                )
                tracer.add_agent_attributes_to_span(span)
                set_span_attribute(
                    span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                )
                model_name = kwargs.get("model", "")

                # Add provider prefix for Groq clients
                if HAS_GROQ:
                    from judgeval.tracer.llm.providers import groq_Groq, groq_AsyncGroq

                    if (
                        isinstance(client, (groq_Groq, groq_AsyncGroq))
                        and model_name
                        and not model_name.startswith("groq/")
                    ):
                        model_name = "groq/" + model_name

                response = function(*args, **kwargs)
                return TracedGenerator(tracer, response, client, span, model_name)
            else:
                with sync_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    tracer.add_agent_attributes_to_span(span)
                    set_span_attribute(
                        span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                    )
                    try:
                        response = function(*args, **kwargs)
                        output, usage = _format_output_data(client, response)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_COMPLETION, output
                        )
                        if usage:
                            _set_usage_attributes(span, usage, tracer)
                        return response
                    except Exception as e:
                        span.record_exception(e)
                        raise

        return wrapper

    def wrapped_async(function, span_name):
        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            if kwargs.get("stream", False):
                span = tracer.get_tracer().start_span(
                    span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                )
                tracer.add_agent_attributes_to_span(span)
                set_span_attribute(
                    span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                )
                model_name = kwargs.get("model", "")

                # Add provider prefix for Groq clients
                if HAS_GROQ:
                    from judgeval.tracer.llm.providers import groq_Groq, groq_AsyncGroq

                    if (
                        isinstance(client, (groq_Groq, groq_AsyncGroq))
                        and model_name
                        and not model_name.startswith("groq/")
                    ):
                        model_name = "groq/" + model_name

                response = await function(*args, **kwargs)
                return TracedAsyncGenerator(tracer, response, client, span, model_name)
            else:
                async with async_span_context(
                    tracer, span_name, {AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
                ) as span:
                    tracer.add_agent_attributes_to_span(span)
                    set_span_attribute(
                        span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
                    )
                    try:
                        response = await function(*args, **kwargs)
                        output, usage = _format_output_data(client, response)
                        set_span_attribute(
                            span, AttributeKeys.GEN_AI_COMPLETION, output
                        )
                        if usage:
                            _set_usage_attributes(span, usage, tracer)
                        return response
                    except Exception as e:
                        span.record_exception(e)
                        raise

        return wrapper

    def wrapped_sync_context_manager(function, span_name):
        """Special wrapper for sync context manager methods."""

        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            span = tracer.get_tracer().start_span(
                span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
            )
            tracer.add_agent_attributes_to_span(span)
            set_span_attribute(
                span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
            )
            model_name = kwargs.get("model", "")

            # Add provider prefix for Groq clients
            if HAS_GROQ:
                from judgeval.tracer.llm.providers import groq_Groq, groq_AsyncGroq

                if (
                    isinstance(client, (groq_Groq, groq_AsyncGroq))
                    and model_name
                    and not model_name.startswith("groq/")
                ):
                    model_name = "groq/" + model_name

            original_context_manager = function(*args, **kwargs)
            return TracedSyncContextManager(
                tracer, original_context_manager, client, span, model_name
            )

        return wrapper

    def wrapped_async_context_manager(function, span_name):
        """Special wrapper for async context manager methods."""

        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            span = tracer.get_tracer().start_span(
                span_name, attributes={AttributeKeys.JUDGMENT_SPAN_KIND: "llm"}
            )
            tracer.add_agent_attributes_to_span(span)
            set_span_attribute(
                span, AttributeKeys.GEN_AI_PROMPT, safe_serialize(kwargs)
            )
            model_name = kwargs.get("model", "")

            # Add provider prefix for Groq clients
            if HAS_GROQ:
                from judgeval.tracer.llm.providers import groq_Groq, groq_AsyncGroq

                if (
                    isinstance(client, (groq_Groq, groq_AsyncGroq))
                    and model_name
                    and not model_name.startswith("groq/")
                ):
                    model_name = "groq/" + model_name

            original_context_manager = function(*args, **kwargs)
            return TracedAsyncContextManager(
                tracer, original_context_manager, client, span, model_name
            )

        return wrapper

    if HAS_OPENAI:
        from judgeval.tracer.llm.providers import openai_OpenAI, openai_AsyncOpenAI

        assert openai_OpenAI is not None, "OpenAI client not found"
        assert openai_AsyncOpenAI is not None, "OpenAI async client not found"
        span_name = "OPENAI_API_CALL"
        if isinstance(client, openai_OpenAI):
            setattr(
                client.chat.completions,
                "create",
                wrapped(client.chat.completions.create, span_name),
            )
            setattr(
                client.responses, "create", wrapped(client.responses.create, span_name)
            )
            setattr(
                client.beta.chat.completions,
                "parse",
                wrapped(client.beta.chat.completions.parse, span_name),
            )
        elif isinstance(client, openai_AsyncOpenAI):
            setattr(
                client.chat.completions,
                "create",
                wrapped_async(client.chat.completions.create, span_name),
            )
            setattr(
                client.responses,
                "create",
                wrapped_async(client.responses.create, span_name),
            )
            setattr(
                client.beta.chat.completions,
                "parse",
                wrapped_async(client.beta.chat.completions.parse, span_name),
            )

    if HAS_TOGETHER:
        from judgeval.tracer.llm.providers import (
            together_Together,
            together_AsyncTogether,
        )

        assert together_Together is not None, "Together client not found"
        assert together_AsyncTogether is not None, "Together async client not found"
        span_name = "TOGETHER_API_CALL"
        if isinstance(client, together_Together):
            setattr(
                client.chat.completions,
                "create",
                wrapped(client.chat.completions.create, span_name),
            )
        elif isinstance(client, together_AsyncTogether):
            setattr(
                client.chat.completions,
                "create",
                wrapped_async(client.chat.completions.create, span_name),
            )

    if HAS_ANTHROPIC:
        from judgeval.tracer.llm.providers import (
            anthropic_Anthropic,
            anthropic_AsyncAnthropic,
        )

        assert anthropic_Anthropic is not None, "Anthropic client not found"
        assert anthropic_AsyncAnthropic is not None, "Anthropic async client not found"
        span_name = "ANTHROPIC_API_CALL"
        if isinstance(client, anthropic_Anthropic):
            setattr(
                client.messages, "create", wrapped(client.messages.create, span_name)
            )
            setattr(
                client.messages,
                "stream",
                wrapped_sync_context_manager(client.messages.stream, span_name),
            )
        elif isinstance(client, anthropic_AsyncAnthropic):
            setattr(
                client.messages,
                "create",
                wrapped_async(client.messages.create, span_name),
            )
            setattr(
                client.messages,
                "stream",
                wrapped_async_context_manager(client.messages.stream, span_name),
            )

    if HAS_GOOGLE_GENAI:
        from judgeval.tracer.llm.providers import (
            google_genai_Client,
            google_genai_AsyncClient,
        )

        assert google_genai_Client is not None, "Google GenAI client not found"
        assert google_genai_AsyncClient is not None, (
            "Google GenAI async client not found"
        )
        span_name = "GOOGLE_API_CALL"
        if isinstance(client, google_genai_Client):
            setattr(
                client.models,
                "generate_content",
                wrapped(client.models.generate_content, span_name),
            )
        elif isinstance(client, google_genai_AsyncClient):
            setattr(
                client.models,
                "generate_content",
                wrapped_async(client.models.generate_content, span_name),
            )

    if HAS_GROQ:
        from judgeval.tracer.llm.providers import groq_Groq, groq_AsyncGroq

        assert groq_Groq is not None, "Groq client not found"
        assert groq_AsyncGroq is not None, "Groq async client not found"
        span_name = "GROQ_API_CALL"
        if isinstance(client, groq_Groq):
            setattr(
                client.chat.completions,
                "create",
                wrapped(client.chat.completions.create, span_name),
            )
        elif isinstance(client, groq_AsyncGroq):
            setattr(
                client.chat.completions,
                "create",
                wrapped_async(client.chat.completions.create, span_name),
            )

    return client


def _format_output_data(
    client: ApiClient, response: Any
) -> tuple[Optional[str], Optional[TraceUsage]]:
    """Format output data from LLM response based on provider."""
    provider_type = _detect_provider(client)

    if provider_type == ProviderType.OPENAI:
        return _format_openai_output(response)
    elif provider_type == ProviderType.ANTHROPIC:
        return _format_anthropic_output(response)
    elif provider_type == ProviderType.TOGETHER:
        return _format_together_output(response)
    elif provider_type == ProviderType.GOOGLE:
        return _format_google_output(response)
    elif provider_type == ProviderType.GROQ:
        return _format_groq_output(response)
    else:
        # Default case - assume OpenAI-compatible for unknown providers
        judgeval_logger.info(
            f"Unknown client type {type(client)}, assuming OpenAI-compatible"
        )
        return _format_openai_output(response)


def _create_usage(
    model_name: str,
    prompt_tokens: int,
    completion_tokens: int,
    cache_read_input_tokens: int = 0,
    cache_creation_input_tokens: int = 0,
) -> TraceUsage:
    prompt_cost, completion_cost = cost_per_token(
        model=model_name,
        prompt_tokens=prompt_tokens,
        completion_tokens=completion_tokens,
        cache_read_input_tokens=cache_read_input_tokens,
        cache_creation_input_tokens=cache_creation_input_tokens,
    )
    total_cost_usd = (
        (prompt_cost + completion_cost) if prompt_cost and completion_cost else None
    )
    return TraceUsage(
        prompt_tokens=prompt_tokens,
        completion_tokens=completion_tokens,
        total_tokens=prompt_tokens + completion_tokens,
        cache_read_input_tokens=cache_read_input_tokens,
        cache_creation_input_tokens=cache_creation_input_tokens,
        prompt_tokens_cost_usd=prompt_cost,
        completion_tokens_cost_usd=completion_cost,
        total_cost_usd=total_cost_usd,
        model_name=model_name,
    )



import datetime
import orjson
import os
import yaml
from dataclasses import dataclass
from typing import List, Literal, Optional

from judgeval.data import Example, Trace
from judgeval.utils.file_utils import get_examples_from_yaml, get_examples_from_json
from judgeval.api import JudgmentSyncClient
from judgeval.logger import judgeval_logger
from judgeval.env import JUDGMENT_API_KEY, JUDGMENT_ORG_ID


@dataclass
class Dataset:
    examples: List[Example]
    traces: List[Trace]
    name: str
    project_name: str
    judgment_api_key: str = JUDGMENT_API_KEY or ""
    organization_id: str = JUDGMENT_ORG_ID or ""

    @classmethod
    def get(
        cls,
        name: str,
        project_name: str,
    ):
        client = JudgmentSyncClient(cls.judgment_api_key, cls.organization_id)
        dataset = client.datasets_pull_for_judgeval(
            {
                "dataset_alias": name,
                "project_name": project_name,
            },
        )
        if not dataset:
            raise ValueError(f"Dataset {name} not found in project {project_name}")
        examples = dataset.get("examples", [])
        for e in examples:
            if isinstance(e, dict) and isinstance(e.get("data"), dict):
                e.update(e.pop("data"))
        judgeval_logger.info(f"Succesfully retrieved dataset {name}!")
        return cls(
            name=name,
            project_name=project_name,
            examples=[Example(**e) for e in examples],
            traces=[Trace(**t) for t in dataset.get("traces", [])],
        )

    @classmethod
    def create(
        cls,
        name: str,
        project_name: str,
        examples: Optional[List[Example]] = None,
        traces: Optional[List[Trace]] = None,
        overwrite: bool = False,
    ):
        if examples and traces:
            raise ValueError("Only one of examples or traces must be provided")

        if not examples:
            examples = []

        if not traces:
            traces = []

        client = JudgmentSyncClient(cls.judgment_api_key, cls.organization_id)
        client.datasets_push(
            {
                "dataset_alias": name,
                "project_name": project_name,
                "examples": [e.model_dump() for e in examples],  # type: ignore
                "traces": [t.model_dump() for t in traces],  # type: ignore
                "overwrite": overwrite,
            }
        )

        judgeval_logger.info(f"Succesfull created dataset {name}!")
        return cls(
            name=name,
            project_name=project_name,
            examples=examples,
            traces=traces,
        )

    def add_from_json(self, file_path: str) -> None:
        """
        Adds examples from a JSON file.

        The JSON file is expected to have the following format:
        [
            {
                "key_01": "value_01",
                "key_02": "value_02"
            },
            {
                "key_11": "value_11",
                "key_12": "value_12",
                "key_13": "value_13"
            },
            ...
        ]
        """
        examples = get_examples_from_json(file_path)
        self.add_examples(examples)

    def add_from_yaml(self, file_path: str) -> None:
        """
        Adds examples from a YAML file.

        The YAML file is expected to have the following format:
        - key_01: value_01
          key_02: value_02
        - key_11: value_11
          key_12: value_12
          key_13: value_13
        ...
        """

        examples = get_examples_from_yaml(file_path)
        self.add_examples(examples)

    def add_examples(self, examples: List[Example]) -> None:
        client = JudgmentSyncClient(self.judgment_api_key, self.organization_id)
        client.datasets_insert_examples(
            {
                "dataset_alias": self.name,
                "project_name": self.project_name,
                "examples": [
                    {
                        "name": e.name,
                        "created_at": e.created_at,
                        "example_id": e.example_id,
                    }
                    for e in examples
                ],
            }
        )

    def add_traces(self, traces: List[Trace]) -> None:
        client = JudgmentSyncClient(self.judgment_api_key, self.organization_id)
        client.traces_add_to_dataset(
            {
                "dataset_alias": self.name,
                "project_name": self.project_name,
                "traces": [t.model_dump() for t in traces],  # type: ignore
            }
        )

    def save_as(
        self,
        file_type: Literal["json", "yaml"],
        dir_path: str,
        save_name: str | None = None,
    ) -> None:
        """
        Saves the dataset as a file. Save only the examples.

        Args:
            file_type (Literal["json", "csv"]): The file type to save the dataset as.
            dir_path (str): The directory path to save the file to.
            save_name (str, optional): The name of the file to save. Defaults to None.
        """
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        file_name = (
            datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            if save_name is None
            else save_name
        )
        complete_path = os.path.join(dir_path, f"{file_name}.{file_type}")
        if file_type == "json":
            with open(complete_path, "wb") as file:
                file.write(
                    orjson.dumps(
                        {
                            "examples": [e.to_dict() for e in self.examples],
                        },
                        option=orjson.OPT_INDENT_2,
                    )
                )
        elif file_type == "yaml":
            with open(complete_path, "w") as file:
                yaml_data = {
                    "examples": [e.to_dict() for e in self.examples],
                }
                yaml.dump(yaml_data, file, default_flow_style=False)
        else:
            ACCEPTABLE_FILE_TYPES = ["json", "yaml"]
            raise TypeError(
                f"Invalid file type: {file_type}. Please choose from {ACCEPTABLE_FILE_TYPES}"
            )

    def __iter__(self):
        return iter(self.examples)

    def __len__(self):
        return len(self.examples)

    def __str__(self):
        return (
            f"{self.__class__.__name__}("
            f"examples={self.examples}, "
            f"traces={self.traces}, "
            f"name={self.name}"
            f")"
        )



# generated by datamodel-codegen:
#   filename:  .openapi.json
#   timestamp: 2025-08-29T04:49:39+00:00

from __future__ import annotations
from typing import Any, Dict, List, Optional, TypedDict, Union
from typing_extensions import NotRequired


TraceAndSpanId = List


class EvalResultsFetch(TypedDict):
    experiment_run_id: str
    project_name: str


class DatasetFetch(TypedDict):
    dataset_alias: str
    project_name: str


class TraceSave(TypedDict):
    project_name: str
    trace_id: str
    name: str
    created_at: str
    duration: float
    offline_mode: NotRequired[bool]
    has_notification: NotRequired[bool]
    customer_id: NotRequired[Optional[str]]
    tags: NotRequired[List[str]]
    metadata: NotRequired[Dict[str, Any]]
    update_id: NotRequired[int]


class TraceFetch(TypedDict):
    trace_id: str


class TraceAddToDataset(TypedDict):
    trace_id: str
    trace_span_id: str
    dataset_alias: str
    project_name: str


class EvaluationRunsBatchRequest(TypedDict):
    organization_id: str
    evaluation_entries: List[Dict[str, Any]]


class ProjectAdd(TypedDict):
    project_name: str


class ProjectAddResponse(TypedDict):
    project_id: str


class ProjectDeleteFromJudgevalResponse(TypedDict):
    project_name: str


class ProjectDeleteResponse(TypedDict):
    message: str


class ScorerExistsRequest(TypedDict):
    name: str


class ScorerExistsResponse(TypedDict):
    exists: bool


class SavePromptScorerRequest(TypedDict):
    name: str
    prompt: str
    threshold: float
    options: NotRequired[Optional[Dict[str, float]]]
    is_trace: NotRequired[Optional[bool]]


class SavePromptScorerResponse(TypedDict):
    message: str
    name: str


class FetchPromptScorerRequest(TypedDict):
    name: str


class CustomScorerUploadPayload(TypedDict):
    scorer_name: str
    scorer_code: str
    requirements_text: str


class CustomScorerTemplateResponse(TypedDict):
    scorer_name: str
    status: str
    message: str


class ResolveProjectNameRequest(TypedDict):
    project_name: str


class ResolveProjectNameResponse(TypedDict):
    project_id: str


class TraceIdRequest(TypedDict):
    trace_id: str


class SpanScoreRequest(TypedDict):
    span_id: str
    trace_id: str


class BaseScorer(TypedDict):
    score_type: str
    threshold: NotRequired[float]
    name: NotRequired[Optional[str]]
    class_name: NotRequired[Optional[str]]
    score: NotRequired[Optional[float]]
    score_breakdown: NotRequired[Optional[Dict[str, Any]]]
    reason: NotRequired[Optional[str]]
    using_native_model: NotRequired[Optional[bool]]
    success: NotRequired[Optional[bool]]
    model: NotRequired[Optional[str]]
    model_client: NotRequired[Any]
    strict_mode: NotRequired[bool]
    error: NotRequired[Optional[str]]
    additional_metadata: NotRequired[Optional[Dict[str, Any]]]
    user: NotRequired[Optional[str]]
    server_hosted: NotRequired[bool]


class ScorerConfig(TypedDict):
    score_type: str
    name: NotRequired[Optional[str]]
    threshold: NotRequired[float]
    strict_mode: NotRequired[bool]
    required_params: NotRequired[List[str]]
    kwargs: NotRequired[Optional[Dict[str, Any]]]


class Example(TypedDict):
    example_id: str
    created_at: str
    name: NotRequired[Optional[str]]


class ValidationError(TypedDict):
    loc: List[Union[str, int]]
    msg: str
    type: str


class SpanBatchItem(TypedDict):
    span_id: str
    trace_id: str
    function: str
    created_at: NotRequired[Any]
    parent_span_id: NotRequired[Optional[str]]
    span_type: NotRequired[Optional[str]]
    inputs: NotRequired[Optional[Dict[str, Any]]]
    output: NotRequired[Any]
    error: NotRequired[Optional[Dict[str, Any]]]
    usage: NotRequired[Optional[Dict[str, Any]]]
    duration: NotRequired[Optional[float]]
    expected_tools: NotRequired[Optional[List[Dict[str, Any]]]]
    additional_metadata: NotRequired[Optional[Dict[str, Any]]]
    has_evaluation: NotRequired[Optional[bool]]
    agent_name: NotRequired[Optional[str]]
    class_name: NotRequired[Optional[str]]
    state_before: NotRequired[Optional[Dict[str, Any]]]
    state_after: NotRequired[Optional[Dict[str, Any]]]
    span_state: str
    update_id: NotRequired[int]
    queued_at: float


class PromptScorer(TypedDict):
    name: str
    prompt: str
    threshold: float
    options: NotRequired[Optional[Dict[str, float]]]
    created_at: NotRequired[Optional[str]]
    updated_at: NotRequired[Optional[str]]
    is_trace: NotRequired[Optional[bool]]


class ScorerData(TypedDict):
    name: str
    threshold: float
    success: bool
    score: NotRequired[Optional[float]]
    reason: NotRequired[Optional[str]]
    strict_mode: NotRequired[Optional[bool]]
    evaluation_model: NotRequired[Union[List[str], str]]
    error: NotRequired[Optional[str]]
    additional_metadata: NotRequired[Optional[Dict[str, Any]]]


class TraceUsage(TypedDict):
    prompt_tokens: NotRequired[Optional[int]]
    completion_tokens: NotRequired[Optional[int]]
    cache_creation_input_tokens: NotRequired[Optional[int]]
    cache_read_input_tokens: NotRequired[Optional[int]]
    total_tokens: NotRequired[Optional[int]]
    prompt_tokens_cost_usd: NotRequired[Optional[float]]
    completion_tokens_cost_usd: NotRequired[Optional[float]]
    total_cost_usd: NotRequired[Optional[float]]
    model_name: NotRequired[Optional[str]]


class Tool(TypedDict):
    tool_name: str
    parameters: NotRequired[Optional[Dict[str, Any]]]
    agent_name: NotRequired[Optional[str]]
    result_dependencies: NotRequired[Optional[List[Dict[str, Any]]]]
    action_dependencies: NotRequired[Optional[List[Dict[str, Any]]]]
    require_all: NotRequired[Optional[bool]]


class ExampleEvaluationRun(TypedDict):
    id: NotRequired[Optional[str]]
    project_name: NotRequired[Optional[str]]
    eval_name: NotRequired[Optional[str]]
    custom_scorers: NotRequired[List[BaseScorer]]
    judgment_scorers: NotRequired[List[ScorerConfig]]
    model: str
    created_at: NotRequired[Optional[str]]
    examples: List[Example]
    trace_span_id: NotRequired[Optional[str]]
    trace_id: NotRequired[Optional[str]]


class HTTPValidationError(TypedDict):
    detail: NotRequired[List[ValidationError]]


class TraceEvaluationRun(TypedDict):
    id: NotRequired[Optional[str]]
    project_name: NotRequired[Optional[str]]
    eval_name: NotRequired[Optional[str]]
    custom_scorers: NotRequired[List[BaseScorer]]
    judgment_scorers: NotRequired[List[ScorerConfig]]
    model: str
    created_at: NotRequired[Optional[str]]
    trace_and_span_ids: List[TraceAndSpanId]
    is_offline: NotRequired[bool]


class DatasetInsertExamples(TypedDict):
    dataset_alias: str
    examples: List[Example]
    project_name: str


class SpansBatchRequest(TypedDict):
    spans: List[SpanBatchItem]
    organization_id: str


class FetchPromptScorerResponse(TypedDict):
    scorer: PromptScorer


class TraceSpan(TypedDict):
    span_id: str
    trace_id: str
    function: str
    created_at: NotRequired[Any]
    parent_span_id: NotRequired[Optional[str]]
    span_type: NotRequired[Optional[str]]
    inputs: NotRequired[Optional[Dict[str, Any]]]
    error: NotRequired[Optional[Dict[str, Any]]]
    output: NotRequired[Any]
    usage: NotRequired[Optional[TraceUsage]]
    duration: NotRequired[Optional[float]]
    expected_tools: NotRequired[Optional[List[Tool]]]
    additional_metadata: NotRequired[Optional[Dict[str, Any]]]
    has_evaluation: NotRequired[Optional[bool]]
    agent_name: NotRequired[Optional[str]]
    class_name: NotRequired[Optional[str]]
    state_before: NotRequired[Optional[Dict[str, Any]]]
    state_after: NotRequired[Optional[Dict[str, Any]]]
    update_id: NotRequired[int]


class Trace(TypedDict):
    trace_id: str
    name: str
    created_at: str
    duration: float
    trace_spans: List[TraceSpan]
    offline_mode: NotRequired[bool]
    rules: NotRequired[Dict[str, Any]]
    has_notification: NotRequired[bool]
    customer_id: NotRequired[Optional[str]]
    tags: NotRequired[List[str]]
    metadata: NotRequired[Dict[str, Any]]
    update_id: NotRequired[int]


class ScoringResult(TypedDict):
    success: bool
    scorers_data: Optional[List[ScorerData]]
    name: NotRequired[Optional[str]]
    data_object: NotRequired[Optional[Union[TraceSpan, Example]]]
    trace_id: NotRequired[Optional[str]]
    run_duration: NotRequired[Optional[float]]
    evaluation_cost: NotRequired[Optional[float]]


class TraceRun(TypedDict):
    project_name: NotRequired[Optional[str]]
    eval_name: NotRequired[Optional[str]]
    traces: List[Trace]
    scorers: List[ScorerConfig]
    model: str
    trace_span_id: NotRequired[Optional[str]]
    tools: NotRequired[Optional[List[Dict[str, Any]]]]


class EvalResults(TypedDict):
    results: List[ScoringResult]
    run: Union[ExampleEvaluationRun, TraceEvaluationRun]


class DatasetPush(TypedDict):
    dataset_alias: str
    comments: NotRequired[Optional[str]]
    source_file: NotRequired[Optional[str]]
    examples: NotRequired[Optional[List[Example]]]
    traces: NotRequired[Optional[List[Trace]]]
    is_trace: NotRequired[bool]
    project_name: str
    overwrite: NotRequired[Optional[bool]]



from judgeval.scorers.base_scorer import BaseScorer
from judgeval.data.judgment_types import Trace as JudgmentTrace
from typing import List, Optional
from abc import abstractmethod


class TraceScorer(BaseScorer):
    @abstractmethod
    async def a_score_trace(
        self, trace: JudgmentTrace, tools: Optional[List] = None, *args, **kwargs
    ) -> float:
        """
        Asynchronously measures the score on a trace
        """
        raise NotImplementedError(
            "You must implement the `a_score_trace` method in your custom scorer"
        )



"""
Util functions for Scorer objects
"""

from typing import List

from judgeval.scorers import BaseScorer


def clone_scorers(scorers: List[BaseScorer]) -> List[BaseScorer]:
    """
    Creates duplicates of the scorers passed as argument.
    """
    cloned_scorers = []
    for s in scorers:
        cloned_scorers.append(s.model_copy(deep=True))
    return cloned_scorers



"""
`judgeval` answer relevancy scorer

TODO add link to docs page for this scorer

"""

# Internal imports
from judgeval.scorers.api_scorer import APIScorerConfig
from judgeval.constants import APIScorerType


class DerailmentScorer(APIScorerConfig):
    score_type: APIScorerType = APIScorerType.DERAILMENT



"""
`judgeval` tool order scorer
"""

# Internal imports
from judgeval.scorers.trace_api_scorer import TraceAPIScorerConfig
from judgeval.constants import APIScorerType
from typing import Dict, Any


class ToolOrderScorer(TraceAPIScorerConfig):
    score_type: APIScorerType = APIScorerType.TOOL_ORDER
    threshold: float = 1.0
    exact_match: bool = False

    def model_dump(self, *args, **kwargs) -> Dict[str, Any]:
        base = super().model_dump(*args, **kwargs)
        base_fields = set(TraceAPIScorerConfig.model_fields.keys())
        all_fields = set(self.__class__.model_fields.keys())

        extra_fields = all_fields - base_fields - {"kwargs"}

        base["kwargs"] = {
            k: getattr(self, k) for k in extra_fields if getattr(self, k) is not None
        }

        return base



"""
Implementation of the ScorerData class.

ScorerData holds the information related to a single, completed Scorer evaluation run.
"""

from __future__ import annotations

from judgeval.data.judgment_types import ScorerData as JudgmentScorerData
from judgeval.scorers import BaseScorer
from typing import List


class ScorerData(JudgmentScorerData):
    """
    ScorerData holds the information related to a single, completed Scorer evaluation run.

    For example, if running the Judgment Faithfulness scorer on an example, the ScorerData
    object will contain whether the example passed its threshold expectation, as well as more detailed
    information surrounding the evaluation run such as the claims and verdicts generated by the
    judge model(s).
    """

    def to_dict(self) -> dict:
        """Convert the ScorerData instance to a JSON-serializable dictionary."""
        return {
            "name": self.name,
            "threshold": self.threshold,
            "success": self.success,
            "score": self.score,
            "reason": self.reason,
            "strict_mode": self.strict_mode,
            "evaluation_model": self.evaluation_model,
            "error": self.error,
            "additional_metadata": self.additional_metadata,
        }


def create_scorer_data(scorer: BaseScorer) -> List[ScorerData]:
    """
    After a `scorer` is run, it contains information about the example that was evaluated
    using the scorer. For example, after computing Faithfulness, the `scorer` object will contain
    whether the example passed its threshold, the score, the reason for score, etc.

    This function takes an executed `scorer` object and produces a ScorerData object that
    contains the output of the scorer run that can be exported to be logged as a part of
    the ScorerResult.
    """
    scorers_result = list()

    scorers_result.append(
        ScorerData(
            name=scorer.name,
            threshold=scorer.threshold,
            score=scorer.score,
            reason=scorer.reason,
            success=scorer.success,
            strict_mode=scorer.strict_mode,
            evaluation_model=scorer.model,
            error=scorer.error,
            additional_metadata=scorer.additional_metadata,
        )
    )
    if hasattr(scorer, "internal_scorer") and scorer.internal_scorer is not None:
        scorers_result.append(
            ScorerData(
                name=scorer.internal_scorer.name,
                score=scorer.internal_scorer.score,
                threshold=scorer.internal_scorer.threshold,
                reason=scorer.internal_scorer.reason,
                success=scorer.internal_scorer.success,
                strict_mode=scorer.internal_scorer.strict_mode,
                evaluation_model=scorer.internal_scorer.model,
                error=scorer.internal_scorer.error,
                additional_metadata=scorer.internal_scorer.additional_metadata,
            )
        )
    return scorers_result



from judgeval.scorers.api_scorer import APIScorerConfig
from judgeval.scorers.base_scorer import BaseScorer
from judgeval.scorers.judgeval_scorers.api_scorers import (
    FaithfulnessScorer,
    AnswerRelevancyScorer,
    AnswerCorrectnessScorer,
    InstructionAdherenceScorer,
    DerailmentScorer,
    ToolOrderScorer,
    PromptScorer,
    ToolDependencyScorer,
)

__all__ = [
    "APIScorerConfig",
    "BaseScorer",
    "PromptScorer",
    "FaithfulnessScorer",
    "AnswerRelevancyScorer",
    "AnswerCorrectnessScorer",
    "InstructionAdherenceScorer",
    "DerailmentScorer",
    "ToolOrderScorer",
    "ToolDependencyScorer",
]



"""
Identifiers used by Judgeval to store specific types of data in the spans.
"""

from opentelemetry.semconv.resource import ResourceAttributes
from opentelemetry.semconv._incubating.attributes import gen_ai_attributes
from enum import Enum


class AttributeKeys(str, Enum):
    # General function tracing attributes (custom namespace)
    JUDGMENT_SPAN_KIND = "judgment.span_kind"
    JUDGMENT_INPUT = "judgment.input"
    JUDGMENT_OUTPUT = "judgment.output"
    JUDGMENT_OFFLINE_MODE = "judgment.offline_mode"
    JUDGMENT_UPDATE_ID = "judgment.update_id"

    # Custom tracking attributes
    JUDGMENT_CUSTOMER_ID = "judgment.customer_id"

    # Agent specific attributes (custom namespace)
    JUDGMENT_AGENT_ID = "judgment.agent_id"
    JUDGMENT_PARENT_AGENT_ID = "judgment.parent_agent_id"
    JUDGMENT_AGENT_CLASS_NAME = "judgment.agent_class_name"
    JUDGMENT_AGENT_INSTANCE_NAME = "judgment.agent_instance_name"
    JUDGMENT_IS_AGENT_ENTRY_POINT = "judgment.is_agent_entry_point"
    JUDGMENT_CUMULATIVE_LLM_COST = "judgment.cumulative_llm_cost"
    JUDGMENT_STATE_BEFORE = "judgment.state_before"
    JUDGMENT_STATE_AFTER = "judgment.state_after"

    # Evaluation-specific attributes (custom namespace)
    PENDING_TRACE_EVAL = "judgment.pending_trace_eval"

    # GenAI-specific attributes (semantic conventions)
    GEN_AI_PROMPT = gen_ai_attributes.GEN_AI_PROMPT
    GEN_AI_COMPLETION = gen_ai_attributes.GEN_AI_COMPLETION
    GEN_AI_REQUEST_MODEL = gen_ai_attributes.GEN_AI_REQUEST_MODEL
    GEN_AI_RESPONSE_MODEL = gen_ai_attributes.GEN_AI_RESPONSE_MODEL
    GEN_AI_SYSTEM = gen_ai_attributes.GEN_AI_SYSTEM
    GEN_AI_USAGE_INPUT_TOKENS = gen_ai_attributes.GEN_AI_USAGE_INPUT_TOKENS
    GEN_AI_USAGE_OUTPUT_TOKENS = gen_ai_attributes.GEN_AI_USAGE_OUTPUT_TOKENS
    GEN_AI_USAGE_COMPLETION_TOKENS = gen_ai_attributes.GEN_AI_USAGE_COMPLETION_TOKENS
    GEN_AI_REQUEST_TEMPERATURE = gen_ai_attributes.GEN_AI_REQUEST_TEMPERATURE
    GEN_AI_REQUEST_MAX_TOKENS = gen_ai_attributes.GEN_AI_REQUEST_MAX_TOKENS
    GEN_AI_RESPONSE_FINISH_REASONS = gen_ai_attributes.GEN_AI_RESPONSE_FINISH_REASONS

    # GenAI-specific attributes (custom namespace)
    GEN_AI_USAGE_TOTAL_COST = "gen_ai.usage.total_cost_usd"


class InternalAttributeKeys(str, Enum):
    """
    Internal attribute keys used for temporary state management in span processors.
    These are NOT exported and are used only for internal span lifecycle management.
    """

    # Span control attributes
    DISABLE_PARTIAL_EMIT = "disable_partial_emit"
    CANCELLED = "cancelled"


class ResourceKeys(str, Enum):
    SERVICE_NAME = ResourceAttributes.SERVICE_NAME
    TELEMETRY_SDK_LANGUAGE = ResourceAttributes.TELEMETRY_SDK_LANGUAGE
    TELEMETRY_SDK_NAME = ResourceAttributes.TELEMETRY_SDK_NAME
    TELEMETRY_SDK_VERSION = ResourceAttributes.TELEMETRY_SDK_VERSION
    JUDGMENT_PROJECT_ID = "judgment.project_id"



from judgeval.scorers.api_scorer import APIScorerConfig
from judgeval.constants import APIScorerType
from typing import Dict, Any, Optional
from judgeval.api import JudgmentSyncClient
from judgeval.exceptions import JudgmentAPIError
import os
from copy import copy
from judgeval.logger import judgeval_logger


def push_prompt_scorer(
    name: str,
    prompt: str,
    threshold: float,
    options: Optional[Dict[str, float]] = None,
    judgment_api_key: str = os.getenv("JUDGMENT_API_KEY") or "",
    organization_id: str = os.getenv("JUDGMENT_ORG_ID") or "",
) -> str:
    client = JudgmentSyncClient(judgment_api_key, organization_id)
    try:
        r = client.save_scorer(
            payload={
                "name": name,
                "prompt": prompt,
                "threshold": threshold,
                "options": options,
            }
        )
    except JudgmentAPIError as e:
        if e.status_code == 500:
            raise JudgmentAPIError(
                status_code=e.status_code,
                detail=f"The server is temporarily unavailable. Please try your request again in a few moments. Error details: {e.detail}",
                response=e.response,
            )
        raise JudgmentAPIError(
            status_code=e.status_code,
            detail=f"Failed to save prompt scorer: {e.detail}",
            response=e.response,
        )
    return r["name"]


def fetch_prompt_scorer(
    name: str,
    judgment_api_key: str = os.getenv("JUDGMENT_API_KEY") or "",
    organization_id: str = os.getenv("JUDGMENT_ORG_ID") or "",
):
    client = JudgmentSyncClient(judgment_api_key, organization_id)
    try:
        scorer_config = client.fetch_scorer({"name": name})["scorer"]
        scorer_config.pop("created_at")
        scorer_config.pop("updated_at")
        return scorer_config
    except JudgmentAPIError as e:
        if e.status_code == 500:
            raise JudgmentAPIError(
                status_code=e.status_code,
                detail=f"The server is temporarily unavailable. Please try your request again in a few moments. Error details: {e.detail}",
                response=e.response,
            )
        raise JudgmentAPIError(
            status_code=e.status_code,
            detail=f"Failed to fetch prompt scorer '{name}': {e.detail}",
            response=e.response,
        )


def scorer_exists(
    name: str,
    judgment_api_key: str = os.getenv("JUDGMENT_API_KEY") or "",
    organization_id: str = os.getenv("JUDGMENT_ORG_ID") or "",
):
    client = JudgmentSyncClient(judgment_api_key, organization_id)
    try:
        return client.scorer_exists({"name": name})["exists"]
    except JudgmentAPIError as e:
        if e.status_code == 500:
            raise JudgmentAPIError(
                status_code=e.status_code,
                detail=f"The server is temporarily unavailable. Please try your request again in a few moments. Error details: {e.detail}",
                response=e.response,
            )
        raise JudgmentAPIError(
            status_code=e.status_code,
            detail=f"Failed to check if scorer exists: {e.detail}",
            response=e.response,
        )


class PromptScorer(APIScorerConfig):
    """
    In the Judgment backend, this scorer is implemented as a PromptScorer that takes
    1. a system role that may involve the Example object
    2. options for scores on the example

    and uses a judge to execute the evaluation from the system role and classify into one of the options
    """

    prompt: str
    options: Optional[Dict[str, float]] = None
    score_type: APIScorerType = APIScorerType.PROMPT_SCORER
    judgment_api_key: str = os.getenv("JUDGMENT_API_KEY") or ""
    organization_id: str = os.getenv("JUDGMENT_ORG_ID") or ""

    @classmethod
    def get(
        cls,
        name: str,
        judgment_api_key: str = os.getenv("JUDGMENT_API_KEY") or "",
        organization_id: str = os.getenv("JUDGMENT_ORG_ID") or "",
    ):
        scorer_config = fetch_prompt_scorer(name, judgment_api_key, organization_id)
        return cls(
            name=name,
            prompt=scorer_config["prompt"],
            threshold=scorer_config["threshold"],
            options=scorer_config.get("options"),
            judgment_api_key=judgment_api_key,
            organization_id=organization_id,
        )

    @classmethod
    def create(
        cls,
        name: str,
        prompt: str,
        threshold: float = 0.5,
        options: Optional[Dict[str, float]] = None,
        judgment_api_key: str = os.getenv("JUDGMENT_API_KEY") or "",
        organization_id: str = os.getenv("JUDGMENT_ORG_ID") or "",
    ):
        if not scorer_exists(name, judgment_api_key, organization_id):
            push_prompt_scorer(
                name, prompt, threshold, options, judgment_api_key, organization_id
            )
            judgeval_logger.info(f"Successfully created PromptScorer: {name}")
            return cls(
                name=name,
                prompt=prompt,
                threshold=threshold,
                options=options,
                judgment_api_key=judgment_api_key,
                organization_id=organization_id,
            )
        else:
            raise JudgmentAPIError(
                status_code=400,
                detail=f"Scorer with name {name} already exists. Either use the existing scorer with the get() method or use a new name.",
                response=None,  # type: ignore
            )

    # Setter functions. Each setter function pushes the scorer to the DB.
    def set_threshold(self, threshold: float):
        """
        Updates the threshold of the scorer.
        """
        self.threshold = threshold
        self.push_prompt_scorer()

    def set_prompt(self, prompt: str):
        """
        Updates the prompt with the new prompt.

        Sample prompt:
        "Did the chatbot answer the user's question in a kind way?"
        """
        self.prompt = prompt
        self.push_prompt_scorer()
        judgeval_logger.info(f"Successfully updated prompt for {self.name}")

    def set_options(self, options: Dict[str, float]):
        """
        Updates the options with the new options.

        Sample options:
        {"yes": 1, "no": 0}
        """
        self.options = options
        self.push_prompt_scorer()
        judgeval_logger.info(f"Successfully updated options for {self.name}")

    def append_to_prompt(self, prompt_addition: str):
        """
        Appends a string to the prompt.
        """
        self.prompt += prompt_addition
        self.push_prompt_scorer()
        judgeval_logger.info(f"Successfully appended to prompt for {self.name}")

    # Getters
    def get_threshold(self) -> float | None:
        """
        Returns the threshold of the scorer.
        """
        return self.threshold

    def get_prompt(self) -> str | None:
        """
        Returns the prompt of the scorer.
        """
        return self.prompt

    def get_options(self) -> Dict[str, float] | None:
        """
        Returns the options of the scorer.
        """
        return copy(self.options) if self.options is not None else None

    def get_name(self) -> str | None:
        """
        Returns the name of the scorer.
        """
        return self.name

    def get_config(self) -> dict:
        """
        Returns a dictionary with all the fields in the scorer.
        """
        return {
            "name": self.name,
            "prompt": self.prompt,
            "threshold": self.threshold,
            "options": self.options,
        }

    def push_prompt_scorer(self):
        """
        Pushes the scorer to the DB.
        """
        push_prompt_scorer(
            self.name,
            self.prompt,
            self.threshold,
            self.options,
            self.judgment_api_key,
            self.organization_id,
        )

    def __str__(self):
        return f"PromptScorer(name={self.name}, prompt={self.prompt}, threshold={self.threshold}, options={self.options})"

    def model_dump(self, *args, **kwargs) -> Dict[str, Any]:
        base = super().model_dump(*args, **kwargs)
        base_fields = set(APIScorerConfig.model_fields.keys())
        all_fields = set(self.__class__.model_fields.keys())

        extra_fields = all_fields - base_fields - {"kwargs"}

        base["kwargs"] = {
            k: getattr(self, k) for k in extra_fields if getattr(self, k) is not None
        }
        return base



from __future__ import annotations
from typing import Any, TypeAlias


HAS_OPENAI = False
openai_OpenAI = None
openai_AsyncOpenAI = None
openai_ChatCompletion = None
openai_Response = None
openai_ParsedChatCompletion = None

try:
    from openai import OpenAI, AsyncOpenAI
    from openai.types.chat.chat_completion import ChatCompletion
    from openai.types.responses.response import Response
    from openai.types.chat import ParsedChatCompletion

    openai_OpenAI = OpenAI
    openai_AsyncOpenAI = AsyncOpenAI
    openai_ChatCompletion = ChatCompletion
    openai_Response = Response
    openai_ParsedChatCompletion = ParsedChatCompletion
    HAS_OPENAI = True
except ImportError:
    pass


HAS_TOGETHER = False
together_Together = None
together_AsyncTogether = None

try:
    from together import Together, AsyncTogether  # type: ignore[import-untyped]

    together_Together = Together
    together_AsyncTogether = AsyncTogether
    HAS_TOGETHER = True
except ImportError:
    pass


HAS_ANTHROPIC = False
anthropic_Anthropic = None
anthropic_AsyncAnthropic = None

try:
    from anthropic import Anthropic, AsyncAnthropic  # type: ignore[import-untyped]

    anthropic_Anthropic = Anthropic
    anthropic_AsyncAnthropic = AsyncAnthropic
    HAS_ANTHROPIC = True
except ImportError:
    pass


HAS_GOOGLE_GENAI = False
google_genai_Client = None
google_genai_cleint_AsyncClient = None

try:
    from google.genai import Client  # type: ignore[import-untyped]
    from google.genai.client import AsyncClient  # type: ignore[import-untyped]

    google_genai_Client = Client
    google_genai_AsyncClient = AsyncClient
    HAS_GOOGLE_GENAI = True
except ImportError:
    pass


HAS_GROQ = False
groq_Groq = None
groq_AsyncGroq = None

try:
    from groq import Groq, AsyncGroq  # type: ignore[import-untyped]

    groq_Groq = Groq
    groq_AsyncGroq = AsyncGroq
    HAS_GROQ = True
except ImportError:
    pass


# TODO: if we support dependency groups we can have this better type, but during runtime, we do
# not know which clients an end user might have installed.
ApiClient: TypeAlias = Any

__all__ = [
    "ApiClient",
    # OpenAI
    "HAS_OPENAI",
    "openai_OpenAI",
    "openai_AsyncOpenAI",
    "openai_ChatCompletion",
    "openai_Response",
    "openai_ParsedChatCompletion",
    # Together
    "HAS_TOGETHER",
    "together_Together",
    "together_AsyncTogether",
    # Anthropic
    "HAS_ANTHROPIC",
    "anthropic_Anthropic",
    "anthropic_AsyncAnthropic",
    # Google GenAI
    "HAS_GOOGLE_GENAI",
    "google_genai_Client",
    "google_genai_AsyncClient",
    # Groq
    "HAS_GROQ",
    "groq_Groq",
    "groq_AsyncGroq",
]



from pydantic import BaseModel
from typing import List, Optional, Dict, Any, Union
from judgeval.data import Trace
from judgeval.scorers import APIScorerConfig, BaseScorer
from judgeval.env import JUDGMENT_DEFAULT_GPT_MODEL


class TraceRun(BaseModel):
    """
    Stores example and evaluation scorers together for running an eval task

    Args:
        project_name (str): The name of the project the evaluation results belong to
        eval_name (str): A name for this evaluation run
        traces (List[Trace]): The traces to evaluate
        scorers (List[Union[JudgmentScorer, BaseScorer]]): A list of scorers to use for evaluation
        model (str): The model used as a judge when using LLM as a Judge
        metadata (Optional[Dict[str, Any]]): Additional metadata to include for this evaluation run, e.g. comments, dataset name, purpose, etc.
        rules (Optional[List[Rule]]): Rules to evaluate against scoring results
        append (Optional[bool]): Whether to append to existing evaluation results
        tools (Optional[List[Dict[str, Any]]]): List of tools to use for evaluation
    """

    organization_id: Optional[str] = None
    project_name: Optional[str] = None
    eval_name: Optional[str] = None
    traces: Optional[List[Trace]] = None
    scorers: List[Union[APIScorerConfig, BaseScorer]]
    model: Optional[str] = JUDGMENT_DEFAULT_GPT_MODEL
    trace_span_id: Optional[str] = None
    append: Optional[bool] = False
    override: Optional[bool] = False

    # TODO: ?
    rules: Any = None
    tools: Optional[List[Dict[str, Any]]] = None

    class Config:
        arbitrary_types_allowed = True



from __future__ import annotations


class SingletonMeta(type):
    """
    Metaclass for creating singleton classes.
    """

    _instances: dict[type, object] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



import sys

if len(sys.argv) != 2:
    print("Usage: python set_version.py <new_version>")
    sys.exit(1)

new_version = sys.argv[1]
version_placeholder = "0.0.0"
found = False

try:
    with open("pyproject.toml", "r") as f:
        lines = f.readlines()
except IOError as e:
    print(f"Error: Failed to read 'pyproject.toml': {e}")
    sys.exit(1)

try:
    with open("pyproject.toml", "w") as f:
        for line in lines:  # Assumes 'lines' was successfully read earlier
            if not found and version_placeholder in line:
                f.write(line.replace(version_placeholder, new_version))
                found = True
            else:
                f.write(line)
except IOError as e:
    print(f"Error: Failed to write to 'pyproject.toml': {e}")
    sys.exit(1)

if not found:
    print("Warning: No '<version_placeholder>' found in pyproject.toml")
    sys.exit(1)



from typing import Any
from opentelemetry.trace import Span
from pydantic import BaseModel
from typing import Callable, Optional
from judgeval.scorers.trace_api_scorer import TraceAPIScorerConfig
from judgeval.env import JUDGMENT_DEFAULT_GPT_MODEL


def set_span_attribute(span: Span, name: str, value: Any):
    if value is None or value == "":
        return

    span.set_attribute(name, value)


class TraceScorerConfig(BaseModel):
    scorer: TraceAPIScorerConfig
    model: str = JUDGMENT_DEFAULT_GPT_MODEL
    sampling_rate: float = 1.0
    run_condition: Optional[Callable[..., bool]] = None



from datetime import datetime, timezone
from judgeval.data.judgment_types import (
    TraceUsage as JudgmentTraceUsage,
    TraceSpan as JudgmentTraceSpan,
    Trace as JudgmentTrace,
)
from judgeval.utils.serialize import json_encoder


class TraceUsage(JudgmentTraceUsage):
    pass


class TraceSpan(JudgmentTraceSpan):
    def model_dump(self, **kwargs):
        return {
            "span_id": self.span_id,
            "trace_id": self.trace_id,
            "created_at": datetime.fromtimestamp(
                self.created_at, tz=timezone.utc
            ).isoformat(),
            "inputs": json_encoder(self.inputs),
            "output": json_encoder(self.output),
            "error": json_encoder(self.error),
            "parent_span_id": self.parent_span_id,
            "function": self.function,
            "duration": self.duration,
            "span_type": self.span_type,
            "usage": self.usage.model_dump() if self.usage else None,
            "has_evaluation": self.has_evaluation,
            "agent_name": self.agent_name,
            "state_before": self.state_before,
            "state_after": self.state_after,
            "additional_metadata": json_encoder(self.additional_metadata),
            "update_id": self.update_id,
        }


class Trace(JudgmentTrace):
    pass



"""
`judgeval` instruction adherence scorer

TODO add link to docs page for this scorer

"""

# Internal imports
from judgeval.scorers.api_scorer import APIScorerConfig
from judgeval.constants import APIScorerType
from judgeval.data import ExampleParams


class InstructionAdherenceScorer(APIScorerConfig):
    def __init__(self, threshold: float):
        super().__init__(
            threshold=threshold,
            score_type=APIScorerType.INSTRUCTION_ADHERENCE,
            required_params=[
                ExampleParams.INPUT,
                ExampleParams.ACTUAL_OUTPUT,
            ],
        )

    @property
    def __name__(self):
        return "Instruction Adherence"



from __future__ import annotations

import time
import uuid
from typing import Any, Dict, List, Optional, Sequence, Set, Type
from uuid import UUID

try:
    from langchain_core.callbacks import BaseCallbackHandler
    from langchain_core.agents import AgentAction, AgentFinish
    from langchain_core.outputs import LLMResult, ChatGeneration
    from langchain_core.messages import (
        AIMessage,
        BaseMessage,
        ChatMessage,
        FunctionMessage,
        HumanMessage,
        SystemMessage,
        ToolMessage,
    )
    from langchain_core.documents import Document
except ImportError as e:
    raise ImportError(
        "Judgeval's langgraph integration requires langchain to be installed. Please install it with `pip install judgeval[langchain]`"
    ) from e

from judgeval.tracer import Tracer
from judgeval.tracer.keys import AttributeKeys
from judgeval.tracer.managers import sync_span_context
from judgeval.utils.serialize import safe_serialize
from judgeval.logger import judgeval_logger
from opentelemetry.trace import Status, StatusCode, Span

# Control flow exception types that should not be treated as errors
CONTROL_FLOW_EXCEPTION_TYPES: Set[Type[BaseException]] = set()

try:
    from langgraph.errors import GraphBubbleUp

    CONTROL_FLOW_EXCEPTION_TYPES.add(GraphBubbleUp)
except ImportError:
    pass

LANGSMITH_TAG_HIDDEN: str = "langsmith:hidden"


class JudgevalCallbackHandler(BaseCallbackHandler):
    """
    LangGraph/LangChain Callback Handler that creates OpenTelemetry spans
    using the Judgeval tracer framework.

    This handler tracks the execution of chains, tools, LLMs, and other components
    in a LangGraph/LangChain application, creating proper span hierarchies for monitoring.
    """

    # Prevent LangChain serialization issues
    lc_serializable = False
    lc_kwargs: dict = {}

    def __init__(self, tracer: Optional[Tracer] = None):
        """
        Initialize the callback handler.

        Args:
            tracer: Optional Tracer instance. If not provided, will try to use an active tracer.
        """
        self.tracer = tracer
        if self.tracer is None:
            # Try to get an active tracer
            if Tracer._active_tracers:
                self.tracer = next(iter(Tracer._active_tracers))
            else:
                judgeval_logger.warning(
                    "No tracer provided and no active tracers found. "
                    "Callback handler will not create spans."
                )
                return

        # Track spans by run_id for proper hierarchy
        self.spans: Dict[UUID, Span] = {}
        self.span_start_times: Dict[UUID, float] = {}
        self.run_id_to_span_id: Dict[UUID, str] = {}
        self.span_id_to_depth: Dict[str, int] = {}
        self.root_run_id: Optional[UUID] = None

        # Track execution for debugging
        self.executed_nodes: List[str] = []
        self.executed_tools: List[str] = []
        self.executed_node_tools: List[Dict[str, Any]] = []

    def reset(self):
        """Reset handler state for reuse across multiple executions."""
        self.spans.clear()
        self.span_start_times.clear()
        self.executed_nodes.clear()
        self.executed_tools.clear()
        self.executed_node_tools.clear()

    def _get_run_name(self, serialized: Optional[Dict[str, Any]], **kwargs: Any) -> str:
        """Extract the name of the operation from serialized data or kwargs."""
        if "name" in kwargs and kwargs["name"] is not None:
            return str(kwargs["name"])

        if serialized is None:
            return "<unknown>"

        try:
            return str(serialized["name"])
        except (KeyError, TypeError):
            pass

        try:
            return str(serialized["id"][-1])
        except (KeyError, TypeError):
            pass

        return "<unknown>"

    def _convert_message_to_dict(self, message: BaseMessage) -> Dict[str, Any]:
        """Convert a LangChain message to a dictionary for storage."""
        if isinstance(message, HumanMessage):
            message_dict = {"role": "user", "content": message.content}
        elif isinstance(message, AIMessage):
            message_dict = {"role": "assistant", "content": message.content}
        elif isinstance(message, SystemMessage):
            message_dict = {"role": "system", "content": message.content}
        elif isinstance(message, ToolMessage):
            message_dict = {
                "role": "tool",
                "content": message.content,
                "tool_call_id": message.tool_call_id,
            }
        elif isinstance(message, FunctionMessage):
            message_dict = {"role": "function", "content": message.content}
        elif isinstance(message, ChatMessage):
            message_dict = {"role": message.role, "content": message.content}
        else:
            message_dict = {"role": "unknown", "content": str(message.content)}

        if hasattr(message, "additional_kwargs") and message.additional_kwargs:
            message_dict["additional_kwargs"] = str(message.additional_kwargs)

        return message_dict

    def _create_message_dicts(
        self, messages: List[BaseMessage]
    ) -> List[Dict[str, Any]]:
        """Convert a list of LangChain messages to dictionaries."""
        return [self._convert_message_to_dict(m) for m in messages]

    def _join_tags_and_metadata(
        self,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        """Join tags and metadata into a single dictionary."""
        final_dict = {}
        if tags is not None and len(tags) > 0:
            final_dict["tags"] = tags
        if metadata is not None:
            final_dict.update(metadata)
        return final_dict if final_dict else None

    def _start_span(
        self,
        run_id: UUID,
        parent_run_id: Optional[UUID],
        name: str,
        span_type: str,
        inputs: Any = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **extra_attributes: Any,
    ) -> None:
        """Start a new span for the given run."""
        if not self.tracer:
            return

        # Skip internal spans
        if name.startswith("__") and name.endswith("__"):
            return

        try:
            # Determine if this is a root span
            is_root = parent_run_id is None
            if is_root:
                self.root_run_id = run_id

            # Calculate depth for proper hierarchy
            current_depth = 0
            if parent_run_id and parent_run_id in self.run_id_to_span_id:
                parent_span_id = self.run_id_to_span_id[parent_run_id]
                current_depth = self.span_id_to_depth.get(parent_span_id, 0) + 1

            # Create span attributes
            attributes = {
                AttributeKeys.JUDGMENT_SPAN_KIND.value: span_type,
            }

            # Add metadata and tags
            combined_metadata = self._join_tags_and_metadata(tags, metadata)
            if combined_metadata:
                metadata_str = safe_serialize(combined_metadata)
                attributes["metadata"] = metadata_str

            # Add extra attributes
            for key, value in extra_attributes.items():
                if value is not None:
                    attributes[str(key)] = str(value)

            # Create span using the tracer's context manager for proper hierarchy
            with sync_span_context(self.tracer, name, attributes) as span:
                # Set input data if provided
                if inputs is not None:
                    span.set_attribute(
                        AttributeKeys.JUDGMENT_INPUT.value, safe_serialize(inputs)
                    )

                # Store span information for tracking
                span_id = (
                    str(span.get_span_context().span_id)
                    if span.get_span_context()
                    else str(uuid.uuid4())
                )
                self.spans[run_id] = span
                self.span_start_times[run_id] = time.time()
                self.run_id_to_span_id[run_id] = span_id
                self.span_id_to_depth[span_id] = current_depth

        except Exception as e:
            judgeval_logger.exception(f"Error starting span for {name}: {e}")

    def _end_span(
        self,
        run_id: UUID,
        outputs: Any = None,
        error: Optional[BaseException] = None,
        **extra_attributes: Any,
    ) -> None:
        """End the span for the given run."""
        if run_id not in self.spans:
            return

        try:
            span = self.spans[run_id]

            # Set output data if provided
            if outputs is not None:
                span.set_attribute(
                    AttributeKeys.JUDGMENT_OUTPUT.value, safe_serialize(outputs)
                )

            # Set additional attributes
            for key, value in extra_attributes.items():
                if value is not None:
                    span.set_attribute(str(key), str(value))

            # Handle errors
            if error is not None:
                # Check if this is a control flow exception
                is_control_flow = any(
                    isinstance(error, t) for t in CONTROL_FLOW_EXCEPTION_TYPES
                )
                if not is_control_flow:
                    span.record_exception(error)
                    span.set_status(Status(StatusCode.ERROR, str(error)))
                # Control flow exceptions don't set error status
            else:
                span.set_status(Status(StatusCode.OK))

            # Note: The span will be ended automatically by the context manager

        except Exception as e:
            judgeval_logger.exception(f"Error ending span for run_id {run_id}: {e}")
        finally:
            # Cleanup tracking data
            if run_id in self.spans:
                del self.spans[run_id]
            if run_id in self.span_start_times:
                del self.span_start_times[run_id]
            if run_id in self.run_id_to_span_id:
                span_id = self.run_id_to_span_id[run_id]
                del self.run_id_to_span_id[run_id]
                if span_id in self.span_id_to_depth:
                    del self.span_id_to_depth[span_id]

            # Check if this is the root run ending
            if run_id == self.root_run_id:
                self.root_run_id = None

    def _log_debug_event(
        self,
        event_name: str,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Log debug information about callback events."""
        judgeval_logger.debug(
            f"Event: {event_name}, run_id: {str(run_id)[:8]}, "
            f"parent_run_id: {str(parent_run_id)[:8] if parent_run_id else None}"
        )

    # Chain callbacks
    def on_chain_start(
        self,
        serialized: Optional[Dict[str, Any]],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a chain starts running."""
        try:
            self._log_debug_event(
                "on_chain_start", run_id, parent_run_id, inputs=inputs
            )

            name = self._get_run_name(serialized, **kwargs)

            # Check for LangGraph node
            node_name = metadata.get("langgraph_node") if metadata else None
            if node_name:
                name = node_name
                if name not in self.executed_nodes:
                    self.executed_nodes.append(name)

            # Determine if this is a root LangGraph execution
            is_langgraph_root = (
                kwargs.get("name") == "LangGraph" and parent_run_id is None
            )
            if is_langgraph_root:
                name = "LangGraph"

            span_level = "DEBUG" if tags and LANGSMITH_TAG_HIDDEN in tags else None

            self._start_span(
                run_id=run_id,
                parent_run_id=parent_run_id,
                name=name,
                span_type="chain",
                inputs=inputs,
                tags=tags,
                metadata=metadata,
                level=span_level,
                serialized=safe_serialize(serialized) if serialized else None,
            )
        except Exception as e:
            judgeval_logger.exception(f"Error in on_chain_start: {e}")

    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a chain ends successfully."""
        try:
            self._log_debug_event(
                "on_chain_end", run_id, parent_run_id, outputs=outputs
            )
            self._end_span(run_id=run_id, outputs=outputs)
        except Exception as e:
            judgeval_logger.exception(f"Error in on_chain_end: {e}")

    def on_chain_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Called when a chain encounters an error."""
        try:
            self._log_debug_event("on_chain_error", run_id, parent_run_id, error=error)
            self._end_span(run_id=run_id, error=error)
        except Exception as e:
            judgeval_logger.exception(f"Error in on_chain_error: {e}")

    # LLM callbacks
    def on_llm_start(
        self,
        serialized: Optional[Dict[str, Any]],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when an LLM starts generating."""
        try:
            self._log_debug_event(
                "on_llm_start", run_id, parent_run_id, prompts=prompts
            )

            name = self._get_run_name(serialized, **kwargs)
            model_name = self._extract_model_name(serialized, kwargs)

            prompt_data = prompts[0] if len(prompts) == 1 else prompts

            self._start_span(
                run_id=run_id,
                parent_run_id=parent_run_id,
                name=name,
                span_type="llm",
                inputs=prompt_data,
                tags=tags,
                metadata=metadata,
                model=model_name,
                serialized=safe_serialize(serialized) if serialized else None,
            )

            # Set GenAI specific attributes
            if run_id in self.spans:
                span = self.spans[run_id]
                if model_name:
                    span.set_attribute(AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)
                span.set_attribute(
                    AttributeKeys.GEN_AI_PROMPT, safe_serialize(prompt_data)
                )

                # Set model parameters if available
                invocation_params = kwargs.get("invocation_params", {})
                if "temperature" in invocation_params:
                    span.set_attribute(
                        AttributeKeys.GEN_AI_REQUEST_TEMPERATURE,
                        float(invocation_params["temperature"]),
                    )
                if "max_tokens" in invocation_params:
                    span.set_attribute(
                        AttributeKeys.GEN_AI_REQUEST_MAX_TOKENS,
                        int(invocation_params["max_tokens"]),
                    )

        except Exception as e:
            judgeval_logger.exception(f"Error in on_llm_start: {e}")

    def on_chat_model_start(
        self,
        serialized: Optional[Dict[str, Any]],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a chat model starts generating."""
        try:
            self._log_debug_event(
                "on_chat_model_start", run_id, parent_run_id, messages=messages
            )

            name = self._get_run_name(serialized, **kwargs)
            model_name = self._extract_model_name(serialized, kwargs)

            # Flatten messages
            flattened_messages = []
            for message_list in messages:
                flattened_messages.extend(self._create_message_dicts(message_list))

            self._start_span(
                run_id=run_id,
                parent_run_id=parent_run_id,
                name=name,
                span_type="llm",
                inputs=flattened_messages,
                tags=tags,
                metadata=metadata,
                model=model_name,
                serialized=safe_serialize(serialized) if serialized else None,
            )

            # Set GenAI specific attributes
            if run_id in self.spans:
                span = self.spans[run_id]
                if model_name:
                    span.set_attribute(AttributeKeys.GEN_AI_REQUEST_MODEL, model_name)
                span.set_attribute(
                    AttributeKeys.GEN_AI_PROMPT, safe_serialize(flattened_messages)
                )

        except Exception as e:
            judgeval_logger.exception(f"Error in on_chat_model_start: {e}")

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when an LLM finishes generating."""
        try:
            self._log_debug_event(
                "on_llm_end", run_id, parent_run_id, response=response
            )

            # Extract response content
            if response.generations:
                last_generation = response.generations[-1][-1]
                if (
                    isinstance(last_generation, ChatGeneration)
                    and last_generation.message
                ):
                    output = self._convert_message_to_dict(last_generation.message)
                else:
                    output = (
                        last_generation.text
                        if hasattr(last_generation, "text")
                        else str(last_generation)
                    )
            else:
                output = ""

            # Extract usage information
            usage_attrs = {}
            if response.llm_output and "token_usage" in response.llm_output:
                token_usage = response.llm_output["token_usage"]
                if hasattr(token_usage, "prompt_tokens"):
                    usage_attrs[AttributeKeys.GEN_AI_USAGE_INPUT_TOKENS] = (
                        token_usage.prompt_tokens
                    )
                if hasattr(token_usage, "completion_tokens"):
                    usage_attrs[AttributeKeys.GEN_AI_USAGE_OUTPUT_TOKENS] = (
                        token_usage.completion_tokens
                    )

            # Set completion attribute
            if run_id in self.spans:
                span = self.spans[run_id]
                span.set_attribute(
                    AttributeKeys.GEN_AI_COMPLETION, safe_serialize(output)
                )

                # Set usage attributes
                for key, value in usage_attrs.items():
                    span.set_attribute(key, value)

            self._end_span(run_id=run_id, outputs=output, **usage_attrs)

        except Exception as e:
            judgeval_logger.exception(f"Error in on_llm_end: {e}")

    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when an LLM encounters an error."""
        try:
            self._log_debug_event("on_llm_error", run_id, parent_run_id, error=error)
            self._end_span(run_id=run_id, error=error)
        except Exception as e:
            judgeval_logger.exception(f"Error in on_llm_error: {e}")

    # Tool callbacks
    def on_tool_start(
        self,
        serialized: Optional[Dict[str, Any]],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a tool starts executing."""
        try:
            self._log_debug_event(
                "on_tool_start", run_id, parent_run_id, input_str=input_str
            )

            name = self._get_run_name(serialized, **kwargs)
            if name not in self.executed_tools:
                self.executed_tools.append(name)

            self._start_span(
                run_id=run_id,
                parent_run_id=parent_run_id,
                name=name,
                span_type="tool",
                inputs=input_str,
                tags=tags,
                metadata=metadata,
                serialized=safe_serialize(serialized) if serialized else None,
            )
        except Exception as e:
            judgeval_logger.exception(f"Error in on_tool_start: {e}")

    def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a tool finishes executing."""
        try:
            self._log_debug_event("on_tool_end", run_id, parent_run_id, output=output)
            self._end_span(run_id=run_id, outputs=output)
        except Exception as e:
            judgeval_logger.exception(f"Error in on_tool_end: {e}")

    def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a tool encounters an error."""
        try:
            self._log_debug_event("on_tool_error", run_id, parent_run_id, error=error)
            self._end_span(run_id=run_id, error=error)
        except Exception as e:
            judgeval_logger.exception(f"Error in on_tool_error: {e}")

    # Agent callbacks
    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when an agent takes an action."""
        try:
            self._log_debug_event(
                "on_agent_action", run_id, parent_run_id, action=action
            )

            if run_id in self.spans:
                span = self.spans[run_id]
                span.set_attribute("agent.action.tool", action.tool)
                span.set_attribute(
                    "agent.action.tool_input", safe_serialize(action.tool_input)
                )
                span.set_attribute("agent.action.log", action.log)

            self._end_span(
                run_id=run_id,
                outputs={"action": action.tool, "input": action.tool_input},
            )
        except Exception as e:
            judgeval_logger.exception(f"Error in on_agent_action: {e}")

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when an agent finishes."""
        try:
            self._log_debug_event(
                "on_agent_finish", run_id, parent_run_id, finish=finish
            )

            if run_id in self.spans:
                span = self.spans[run_id]
                span.set_attribute("agent.finish.log", finish.log)

            self._end_span(run_id=run_id, outputs=finish.return_values)
        except Exception as e:
            judgeval_logger.exception(f"Error in on_agent_finish: {e}")

    # Retriever callbacks
    def on_retriever_start(
        self,
        serialized: Optional[Dict[str, Any]],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a retriever starts."""
        try:
            self._log_debug_event(
                "on_retriever_start", run_id, parent_run_id, query=query
            )

            name = self._get_run_name(serialized, **kwargs)

            self._start_span(
                run_id=run_id,
                parent_run_id=parent_run_id,
                name=name,
                span_type="retriever",
                inputs=query,
                tags=tags,
                metadata=metadata,
                serialized=safe_serialize(serialized) if serialized else None,
            )
        except Exception as e:
            judgeval_logger.exception(f"Error in on_retriever_start: {e}")

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a retriever finishes."""
        try:
            self._log_debug_event(
                "on_retriever_end", run_id, parent_run_id, documents=documents
            )

            # Convert documents to serializable format
            doc_data = [
                {"page_content": doc.page_content, "metadata": doc.metadata}
                for doc in documents
            ]

            if run_id in self.spans:
                span = self.spans[run_id]
                span.set_attribute("retriever.document_count", len(documents))

            self._end_span(
                run_id=run_id, outputs=doc_data, document_count=len(documents)
            )
        except Exception as e:
            judgeval_logger.exception(f"Error in on_retriever_end: {e}")

    def on_retriever_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Called when a retriever encounters an error."""
        try:
            self._log_debug_event(
                "on_retriever_error", run_id, parent_run_id, error=error
            )
            self._end_span(run_id=run_id, error=error)
        except Exception as e:
            judgeval_logger.exception(f"Error in on_retriever_error: {e}")

    def _extract_model_name(
        self, serialized: Optional[Dict[str, Any]], kwargs: Dict[str, Any]
    ) -> Optional[str]:
        """Extract model name from serialized data or kwargs."""
        # Try to get from invocation params
        invocation_params = kwargs.get("invocation_params", {})
        if "model_name" in invocation_params:
            return invocation_params["model_name"]
        if "model" in invocation_params:
            return invocation_params["model"]

        # Try to get from serialized data
        if serialized:
            if "model_name" in serialized:
                return serialized["model_name"]
            if "model" in serialized:
                return serialized["model"]

        return None


__all__ = ["JudgevalCallbackHandler"]



# generated by datamodel-codegen:
#   filename:  .openapi.json
#   timestamp: 2025-08-29T04:49:38+00:00

from __future__ import annotations
from typing import Annotated, Any, Dict, List, Optional, Union
from pydantic import AwareDatetime, BaseModel, ConfigDict, Field, RootModel


class TraceAndSpanId(RootModel[List]):
    root: Annotated[List, Field(max_length=2, min_length=2)]


class EvalResultsFetch(BaseModel):
    experiment_run_id: Annotated[str, Field(title="Experiment Run Id")]
    project_name: Annotated[str, Field(title="Project Name")]


class DatasetFetch(BaseModel):
    dataset_alias: Annotated[str, Field(title="Dataset Alias")]
    project_name: Annotated[str, Field(title="Project Name")]


class TraceSave(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]
    trace_id: Annotated[str, Field(title="Trace Id")]
    name: Annotated[str, Field(title="Name")]
    created_at: Annotated[str, Field(title="Created At")]
    duration: Annotated[float, Field(title="Duration")]
    offline_mode: Annotated[Optional[bool], Field(title="Offline Mode")] = False
    has_notification: Annotated[Optional[bool], Field(title="Has Notification")] = False
    customer_id: Annotated[Optional[str], Field(title="Customer Id")] = None
    tags: Annotated[Optional[List[str]], Field(title="Tags")] = None
    metadata: Annotated[Optional[Dict[str, Any]], Field(title="Metadata")] = None
    update_id: Annotated[Optional[int], Field(title="Update Id")] = 1


class TraceFetch(BaseModel):
    trace_id: Annotated[str, Field(title="Trace Id")]


class TraceAddToDataset(BaseModel):
    trace_id: Annotated[str, Field(title="Trace Id")]
    trace_span_id: Annotated[str, Field(title="Trace Span Id")]
    dataset_alias: Annotated[str, Field(title="Dataset Alias")]
    project_name: Annotated[str, Field(title="Project Name")]


class EvaluationRunsBatchRequest(BaseModel):
    organization_id: Annotated[str, Field(title="Organization Id")]
    evaluation_entries: Annotated[
        List[Dict[str, Any]], Field(title="Evaluation Entries")
    ]


class ProjectAdd(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]


class ProjectAddResponse(BaseModel):
    project_id: Annotated[str, Field(title="Project Id")]


class ProjectDeleteFromJudgevalResponse(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]


class ProjectDeleteResponse(BaseModel):
    message: Annotated[str, Field(title="Message")]


class ScorerExistsRequest(BaseModel):
    name: Annotated[str, Field(title="Name")]


class ScorerExistsResponse(BaseModel):
    exists: Annotated[bool, Field(title="Exists")]


class SavePromptScorerRequest(BaseModel):
    name: Annotated[str, Field(title="Name")]
    prompt: Annotated[str, Field(title="Prompt")]
    threshold: Annotated[float, Field(title="Threshold")]
    options: Annotated[Optional[Dict[str, float]], Field(title="Options")] = None
    is_trace: Annotated[Optional[bool], Field(title="Is Trace")] = False


class SavePromptScorerResponse(BaseModel):
    message: Annotated[str, Field(title="Message")]
    name: Annotated[str, Field(title="Name")]


class FetchPromptScorerRequest(BaseModel):
    name: Annotated[str, Field(title="Name")]


class CustomScorerUploadPayload(BaseModel):
    scorer_name: Annotated[str, Field(title="Scorer Name")]
    scorer_code: Annotated[str, Field(title="Scorer Code")]
    requirements_text: Annotated[str, Field(title="Requirements Text")]


class CustomScorerTemplateResponse(BaseModel):
    scorer_name: Annotated[str, Field(title="Scorer Name")]
    status: Annotated[str, Field(title="Status")]
    message: Annotated[str, Field(title="Message")]


class ResolveProjectNameRequest(BaseModel):
    project_name: Annotated[str, Field(title="Project Name")]


class ResolveProjectNameResponse(BaseModel):
    project_id: Annotated[str, Field(title="Project Id")]


class TraceIdRequest(BaseModel):
    trace_id: Annotated[str, Field(title="Trace Id")]


class SpanScoreRequest(BaseModel):
    span_id: Annotated[str, Field(title="Span Id")]
    trace_id: Annotated[str, Field(title="Trace Id")]


class BaseScorer(BaseModel):
    score_type: Annotated[str, Field(title="Score Type")]
    threshold: Annotated[Optional[float], Field(title="Threshold")] = 0.5
    name: Annotated[Optional[str], Field(title="Name")] = None
    class_name: Annotated[Optional[str], Field(title="Class Name")] = None
    score: Annotated[Optional[float], Field(title="Score")] = None
    score_breakdown: Annotated[
        Optional[Dict[str, Any]], Field(title="Score Breakdown")
    ] = None
    reason: Annotated[Optional[str], Field(title="Reason")] = ""
    using_native_model: Annotated[Optional[bool], Field(title="Using Native Model")] = (
        None
    )
    success: Annotated[Optional[bool], Field(title="Success")] = None
    model: Annotated[Optional[str], Field(title="Model")] = None
    model_client: Annotated[Any, Field(title="Model Client")] = None
    strict_mode: Annotated[Optional[bool], Field(title="Strict Mode")] = False
    error: Annotated[Optional[str], Field(title="Error")] = None
    additional_metadata: Annotated[
        Optional[Dict[str, Any]], Field(title="Additional Metadata")
    ] = None
    user: Annotated[Optional[str], Field(title="User")] = None
    server_hosted: Annotated[Optional[bool], Field(title="Server Hosted")] = False


class ScorerConfig(BaseModel):
    score_type: Annotated[str, Field(title="Score Type")]
    name: Annotated[Optional[str], Field(title="Name")] = None
    threshold: Annotated[Optional[float], Field(title="Threshold")] = 0.5
    strict_mode: Annotated[Optional[bool], Field(title="Strict Mode")] = False
    required_params: Annotated[Optional[List[str]], Field(title="Required Params")] = []
    kwargs: Annotated[Optional[Dict[str, Any]], Field(title="Kwargs")] = None


class Example(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    example_id: Annotated[str, Field(title="Example Id")]
    created_at: Annotated[str, Field(title="Created At")]
    name: Annotated[Optional[str], Field(title="Name")] = None


class ValidationError(BaseModel):
    loc: Annotated[List[Union[str, int]], Field(title="Location")]
    msg: Annotated[str, Field(title="Message")]
    type: Annotated[str, Field(title="Error Type")]


class SpanBatchItem(BaseModel):
    span_id: Annotated[str, Field(title="Span Id")]
    trace_id: Annotated[str, Field(title="Trace Id")]
    function: Annotated[str, Field(title="Function")]
    created_at: Annotated[Any, Field(title="Created At")] = None
    parent_span_id: Annotated[Optional[str], Field(title="Parent Span Id")] = None
    span_type: Annotated[Optional[str], Field(title="Span Type")] = "span"
    inputs: Annotated[Optional[Dict[str, Any]], Field(title="Inputs")] = None
    output: Annotated[Any, Field(title="Output")] = None
    error: Annotated[Optional[Dict[str, Any]], Field(title="Error")] = None
    usage: Annotated[Optional[Dict[str, Any]], Field(title="Usage")] = None
    duration: Annotated[Optional[float], Field(title="Duration")] = None
    expected_tools: Annotated[
        Optional[List[Dict[str, Any]]], Field(title="Expected Tools")
    ] = None
    additional_metadata: Annotated[
        Optional[Dict[str, Any]], Field(title="Additional Metadata")
    ] = None
    has_evaluation: Annotated[Optional[bool], Field(title="Has Evaluation")] = False
    agent_name: Annotated[Optional[str], Field(title="Agent Name")] = None
    class_name: Annotated[Optional[str], Field(title="Class Name")] = None
    state_before: Annotated[Optional[Dict[str, Any]], Field(title="State Before")] = (
        None
    )
    state_after: Annotated[Optional[Dict[str, Any]], Field(title="State After")] = None
    span_state: Annotated[str, Field(title="Span State")]
    update_id: Annotated[Optional[int], Field(title="Update Id")] = 1
    queued_at: Annotated[float, Field(title="Queued At")]


class PromptScorer(BaseModel):
    name: Annotated[str, Field(title="Name")]
    prompt: Annotated[str, Field(title="Prompt")]
    threshold: Annotated[float, Field(title="Threshold")]
    options: Annotated[Optional[Dict[str, float]], Field(title="Options")] = None
    created_at: Annotated[Optional[AwareDatetime], Field(title="Created At")] = None
    updated_at: Annotated[Optional[AwareDatetime], Field(title="Updated At")] = None
    is_trace: Annotated[Optional[bool], Field(title="Is Trace")] = False


class ScorerData(BaseModel):
    name: Annotated[str, Field(title="Name")]
    threshold: Annotated[float, Field(title="Threshold")]
    success: Annotated[bool, Field(title="Success")]
    score: Annotated[Optional[float], Field(title="Score")] = None
    reason: Annotated[Optional[str], Field(title="Reason")] = None
    strict_mode: Annotated[Optional[bool], Field(title="Strict Mode")] = None
    evaluation_model: Annotated[
        Optional[Union[List[str], str]], Field(title="Evaluation Model")
    ] = None
    error: Annotated[Optional[str], Field(title="Error")] = None
    additional_metadata: Annotated[
        Optional[Dict[str, Any]], Field(title="Additional Metadata")
    ] = None


class TraceUsage(BaseModel):
    prompt_tokens: Annotated[Optional[int], Field(title="Prompt Tokens")] = None
    completion_tokens: Annotated[Optional[int], Field(title="Completion Tokens")] = None
    cache_creation_input_tokens: Annotated[
        Optional[int], Field(title="Cache Creation Input Tokens")
    ] = None
    cache_read_input_tokens: Annotated[
        Optional[int], Field(title="Cache Read Input Tokens")
    ] = None
    total_tokens: Annotated[Optional[int], Field(title="Total Tokens")] = None
    prompt_tokens_cost_usd: Annotated[
        Optional[float], Field(title="Prompt Tokens Cost Usd")
    ] = None
    completion_tokens_cost_usd: Annotated[
        Optional[float], Field(title="Completion Tokens Cost Usd")
    ] = None
    total_cost_usd: Annotated[Optional[float], Field(title="Total Cost Usd")] = None
    model_name: Annotated[Optional[str], Field(title="Model Name")] = None


class Tool(BaseModel):
    tool_name: Annotated[str, Field(title="Tool Name")]
    parameters: Annotated[Optional[Dict[str, Any]], Field(title="Parameters")] = None
    agent_name: Annotated[Optional[str], Field(title="Agent Name")] = None
    result_dependencies: Annotated[
        Optional[List[Dict[str, Any]]], Field(title="Result Dependencies")
    ] = None
    action_dependencies: Annotated[
        Optional[List[Dict[str, Any]]], Field(title="Action Dependencies")
    ] = None
    require_all: Annotated[Optional[bool], Field(title="Require All")] = None


class ExampleEvaluationRun(BaseModel):
    id: Annotated[Optional[str], Field(title="Id")] = None
    project_name: Annotated[Optional[str], Field(title="Project Name")] = None
    eval_name: Annotated[Optional[str], Field(title="Eval Name")] = None
    custom_scorers: Annotated[
        Optional[List[BaseScorer]], Field(title="Custom Scorers")
    ] = []
    judgment_scorers: Annotated[
        Optional[List[ScorerConfig]], Field(title="Judgment Scorers")
    ] = []
    model: Annotated[str, Field(title="Model")]
    created_at: Annotated[Optional[str], Field(title="Created At")] = None
    examples: Annotated[List[Example], Field(title="Examples")]
    trace_span_id: Annotated[Optional[str], Field(title="Trace Span Id")] = None
    trace_id: Annotated[Optional[str], Field(title="Trace Id")] = None


class HTTPValidationError(BaseModel):
    detail: Annotated[Optional[List[ValidationError]], Field(title="Detail")] = None


class TraceEvaluationRun(BaseModel):
    id: Annotated[Optional[str], Field(title="Id")] = None
    project_name: Annotated[Optional[str], Field(title="Project Name")] = None
    eval_name: Annotated[Optional[str], Field(title="Eval Name")] = None
    custom_scorers: Annotated[
        Optional[List[BaseScorer]], Field(title="Custom Scorers")
    ] = []
    judgment_scorers: Annotated[
        Optional[List[ScorerConfig]], Field(title="Judgment Scorers")
    ] = []
    model: Annotated[str, Field(title="Model")]
    created_at: Annotated[Optional[str], Field(title="Created At")] = None
    trace_and_span_ids: Annotated[
        List[TraceAndSpanId], Field(title="Trace And Span Ids")
    ]
    is_offline: Annotated[Optional[bool], Field(title="Is Offline")] = False


class DatasetInsertExamples(BaseModel):
    dataset_alias: Annotated[str, Field(title="Dataset Alias")]
    examples: Annotated[List[Example], Field(title="Examples")]
    project_name: Annotated[str, Field(title="Project Name")]


class SpansBatchRequest(BaseModel):
    spans: Annotated[List[SpanBatchItem], Field(title="Spans")]
    organization_id: Annotated[str, Field(title="Organization Id")]


class FetchPromptScorerResponse(BaseModel):
    scorer: PromptScorer


class TraceSpan(BaseModel):
    span_id: Annotated[str, Field(title="Span Id")]
    trace_id: Annotated[str, Field(title="Trace Id")]
    function: Annotated[str, Field(title="Function")]
    created_at: Annotated[Any, Field(title="Created At")] = None
    parent_span_id: Annotated[Optional[str], Field(title="Parent Span Id")] = None
    span_type: Annotated[Optional[str], Field(title="Span Type")] = "span"
    inputs: Annotated[Optional[Dict[str, Any]], Field(title="Inputs")] = None
    error: Annotated[Optional[Dict[str, Any]], Field(title="Error")] = None
    output: Annotated[Any, Field(title="Output")] = None
    usage: Optional[TraceUsage] = None
    duration: Annotated[Optional[float], Field(title="Duration")] = None
    expected_tools: Annotated[Optional[List[Tool]], Field(title="Expected Tools")] = (
        None
    )
    additional_metadata: Annotated[
        Optional[Dict[str, Any]], Field(title="Additional Metadata")
    ] = None
    has_evaluation: Annotated[Optional[bool], Field(title="Has Evaluation")] = False
    agent_name: Annotated[Optional[str], Field(title="Agent Name")] = None
    class_name: Annotated[Optional[str], Field(title="Class Name")] = None
    state_before: Annotated[Optional[Dict[str, Any]], Field(title="State Before")] = (
        None
    )
    state_after: Annotated[Optional[Dict[str, Any]], Field(title="State After")] = None
    update_id: Annotated[Optional[int], Field(title="Update Id")] = 1


class Trace(BaseModel):
    trace_id: Annotated[str, Field(title="Trace Id")]
    name: Annotated[str, Field(title="Name")]
    created_at: Annotated[str, Field(title="Created At")]
    duration: Annotated[float, Field(title="Duration")]
    trace_spans: Annotated[List[TraceSpan], Field(title="Trace Spans")]
    offline_mode: Annotated[Optional[bool], Field(title="Offline Mode")] = False
    rules: Annotated[Optional[Dict[str, Any]], Field(title="Rules")] = {}
    has_notification: Annotated[Optional[bool], Field(title="Has Notification")] = False
    customer_id: Annotated[Optional[str], Field(title="Customer Id")] = None
    tags: Annotated[Optional[List[str]], Field(title="Tags")] = []
    metadata: Annotated[Optional[Dict[str, Any]], Field(title="Metadata")] = {}
    update_id: Annotated[Optional[int], Field(title="Update Id")] = 1


class ScoringResult(BaseModel):
    success: Annotated[bool, Field(title="Success")]
    scorers_data: Annotated[Optional[List[ScorerData]], Field(title="Scorers Data")] = (
        None
    )
    name: Annotated[Optional[str], Field(title="Name")] = None
    data_object: Annotated[
        Optional[Union[TraceSpan, Example]], Field(title="Data Object")
    ] = None
    trace_id: Annotated[Optional[str], Field(title="Trace Id")] = None
    run_duration: Annotated[Optional[float], Field(title="Run Duration")] = None
    evaluation_cost: Annotated[Optional[float], Field(title="Evaluation Cost")] = None


class TraceRun(BaseModel):
    project_name: Annotated[Optional[str], Field(title="Project Name")] = None
    eval_name: Annotated[Optional[str], Field(title="Eval Name")] = None
    traces: Annotated[List[Trace], Field(title="Traces")]
    scorers: Annotated[List[ScorerConfig], Field(title="Scorers")]
    model: Annotated[str, Field(title="Model")]
    trace_span_id: Annotated[Optional[str], Field(title="Trace Span Id")] = None
    tools: Annotated[Optional[List[Dict[str, Any]]], Field(title="Tools")] = None


class EvalResults(BaseModel):
    results: Annotated[List[ScoringResult], Field(title="Results")]
    run: Annotated[Union[ExampleEvaluationRun, TraceEvaluationRun], Field(title="Run")]


class DatasetPush(BaseModel):
    dataset_alias: Annotated[str, Field(title="Dataset Alias")]
    comments: Annotated[Optional[str], Field(title="Comments")] = None
    source_file: Annotated[Optional[str], Field(title="Source File")] = None
    examples: Annotated[Optional[List[Example]], Field(title="Examples")] = None
    traces: Annotated[Optional[List[Trace]], Field(title="Traces")] = None
    is_trace: Annotated[Optional[bool], Field(title="Is Trace")] = False
    project_name: Annotated[str, Field(title="Project Name")]
    overwrite: Annotated[Optional[bool], Field(title="Overwrite")] = False



"""
`judgeval` tool dependency scorer
"""

# Internal imports
from judgeval.scorers.trace_api_scorer import TraceAPIScorerConfig
from judgeval.constants import APIScorerType
from typing import Optional, Dict


class ToolDependencyScorer(TraceAPIScorerConfig):
    kwargs: Optional[Dict] = None

    def __init__(self, threshold: float = 1.0, enable_param_checking: bool = True):
        super().__init__(threshold=threshold, score_type=APIScorerType.TOOL_DEPENDENCY)
        self.kwargs = {"enable_param_checking": enable_param_checking}

    @property
    def __name__(self):
        return "Tool Dependency"
